/* --------------------------- 3DSRDR.C -------------------------------

	Program:	3DS reader
	Purpose:	Provide a modular parser for 3DS files
	Author:		Geshem ( erez27@gmail.com )

	Special thanks to:
		http://www.the-labs.com/Blender/3DS-details.html	- for providing me with lotsa info
*/

#include "3dsrdr.h"
#include "assert.h"


/* Print out 3DS information? */
boolean printf_3ds = ttrue;

/* Global variables to hold the current chunk handlers list */
unsigned chunk_handlers_list_len;
chunk_handler* chunk_handlers_list;



/* A list of chunk IDs and their corresponding names */
chunk_name chunk_names_list[] = {
	{CHUNK_M3D_VERSION,	"M3D Version"},
	{CHUNK_RGBF,	"RGB float"},
	{CHUNK_RGBB,	"RGB byte"},
	{CHUNK_MAIN,	"Main"},
	{CHUNK_OBJMESH,	"Object Mesh"},
	{CHUNK_MESH_VERSION,	"Mesh Version"},
	{CHUNK_MASTER_SCALE,	"Master Scale"},
	{CHUNK_BKGCOLOR,	"Background color"},
	{CHUNK_AMBCOLOR,	"Ambient color"},
	{CHUNK_OBJBLOCK,	"Object Block"},
	{CHUNK_TRIMESH,	"Tri-Mesh"},
	{CHUNK_VERTLIST,	"Vertex list"},
	{CHUNK_FACELIST,	"Face list"},
	{CHUNK_FACEMAT,	"Face material"},
	{CHUNK_MAPLIST,	"Mappings list"},
	{CHUNK_SMOOLIST,	"Smoothings"},
	{CHUNK_TRMATRIX,	"Matrix"},
	{CHUNK_LIGHT,	"Light"},
	{CHUNK_SPOTLIGHT,	"Spotlight"},
	{CHUNK_CAMERA,	"Camera"},
	{CHUNK_MATERIAL,	"Material"},
	{CHUNK_MATNAME,	"Material name"},
	{CHUNK_AMBIENT,	"Ambient color"},
	{CHUNK_DIFFUSE,	"Diffuse color"},
	{CHUNK_SPECULAR,	"Specular color"},
	{CHUNK_TEXTURE,	"Texture map"},
	{CHUNK_BUMPMAP,	"Bump map"},
	{CHUNK_MAPFILE,	"Map filename"},
	{CHUNK_KEYFRAMER,	"Keyframer data"},
	{CHUNK_KFHDR,	"Keyframer header"},
	{CHUNK_FRAMES,	"Frames"},
	{CHUNK_KFCURTIME,	"Current Keyframe"},
	{CHUNK_OBJECT_NODE_TAG,	"Object Node Tag"},
	{CHUNK_CAMERA_NODE_TAG,	"Camera Node Tag"},
	{CHUNK_TARGET_NODE_TAG,	"Target Node Tag"},
	{CHUNK_NODE_ID,	"Node ID"},
	{CHUNK_NODE_HDR,	"Node Header"},
	{CHUNK_PIVOT,	"Pivot"},
	{CHUNK_INSTANCE_NAME,	"Instance Name"},
	{CHUNK_POS_TRACK_TAG,	"Position Track Tag"},
	{CHUNK_ROT_TRACK_TAG,	"Rotation Track Tag"},
	{CHUNK_SCL_TRACK_TAG,	"Scale Track Tag"},
	{CHUNK_FOV_TRACK_TAG,	"Camera Field Of View Track Tag"},
	{CHUNK_ROL_TRACK_TAG,	"Camera Roll Track Tag"},
	{CHUNK_BOUNDBOX,	"BoundBox"}
};



void m3d_report_error(FILE* f) {
	printf("Error in pos %ld\n",ftell(f));
}


/* A shared buffer to be used by "m3d_fread_name" to hold the names it reads */
#define NAME_BUFFER_SIZE 256
char name_buffer[NAME_BUFFER_SIZE];

/* Read a C string (ASCIIZ) from file 'f' into global var 'name_buffer'. */
/* Stops reading on null-terminator or end-of-file. */
/* Stops storing if length exceeds maximum (defined by NAME_BUFFER_SIZE) */
void m3d_fread_name(FILE* f) {
	unsigned c=0;
	int i=0;
	while ( (i = fgetc(f)) != EOF && i != (int)'\0') if (c < NAME_BUFFER_SIZE) name_buffer[c++] = (char)i;
	name_buffer[c]='\0';
}


/* ------------------------------------ */


/* Search chunk_names_list for the given id and return it's corresponding name */
char* find_chunk_name(word id) {
    unsigned i;
    for (i = 0; i < (unsigned)(sizeof(chunk_names_list)/sizeof(chunk_name)); i++)
        if (id == chunk_names_list[i].id)
            return chunk_names_list[i].name;
    return NULL;
}

/* Search chunk_handlers_list for the given id and return it's corresponding handler */
chunk_handler_func_t* find_chunk_func(word id) {	/* Should return a real error value... No? */
	unsigned i;

	for (i=0; i<chunk_handlers_list_len; ++i)
		if (id == chunk_handlers_list[i].id)
			return chunk_handlers_list[i].func;
	return SKIP_CHUNK;
}

/* Read a chunk from f and call its appropriate handler */
void chunk_reader(FILE* f, int ind, long p) {
	struct chunk_header h;
	char* name;
	long pc;
	chunk_handler_func_t* func;

	pc = ftell(f);
	while (pc < p) {

		/*if (fread(&h, sizeof(h), 1, f) != 1) return;*/
		m3d_fread_ext(h,f);

		if (printf_3ds) {
			name = find_chunk_name(h.id);
			if (name != NULL)
				printf("%*cChunk type: \"%s\", offset 0x%lX, size %ld bytes\n", ind, ' ', name, pc, h.len);
			else
				printf("%*cUnknown chunk: 0x%04X, offset 0x%lX, size: %ld bytes.\n", ind, ' ', h.id, pc, h.len);
			(void)fflush(stdout);
		}

		pc += h.len;

		func = find_chunk_func(h.id);
		ind += DUMP3DS_INDENT;

		if ( func != SKIP_CHUNK )
			if (func != NULL)
				func(f, ind , pc);
			else
				chunk_reader(f, ind , pc);
		else
			if(printf_3ds) printf("%*c  *** Skipped ***\n", ind, ' ');

		ind -= DUMP3DS_INDENT;

		(void)fseek(f, pc, SEEK_SET);

		if (ferror(f)) {
			if(printf_3ds) printf("Error occured!\n");
			break;
		}
	}
}


/* Start process a 3DS file ("filename") using handlers from handlers_list */
void process_3ds(char* filename, chunk_handler* handlers_list) {
	FILE* f;
	long size;
	unsigned i;

	/*chunk_handlers_list_len = handlers_list_len;*/
	chunk_handlers_list = handlers_list;
	for (i=0;handlers_list[i].id!=CHUNK_NULL;++i);
	assert( handlers_list[i].func == NULL );
	chunk_handlers_list_len = i;
	
	

    if( (f = fopen(filename, "rb")) == NULL ) {
		fprintf(stderr,"3dsrdr: Error opening %s for read. Quitting.",filename);
		return;
	}

    (void)fseek(f, 0, SEEK_END);
    size = ftell(f);
    (void)fseek(f, 0, SEEK_SET);

    chunk_reader(f, 0, size);

    (void)fclose(f);
		
}






/**********************************************************\
|*                                                         |
|*              Print 3DS chunk handlers                   |
|*                                                         |
\**********************************************************/
/* Nicely prints out every information read from the 3DS */

void print3ds_RGBF_reader (FILE* f, int ind, long p) {
    float c[3];
	m3d_fread_ext(c,f);	
    M3D_REPORT_CHUNK_DATA(("Red: %f, Green: %f, Blue: %f", c[0], c[1], c[2]));
}

void print3ds_RGBB_reader (FILE* f, int ind, long p) {
    byte c[3];
	m3d_fread_ext(c,f);	
    M3D_REPORT_CHUNK_DATA(("Red: %d, Green: %d, Blue: %d", c[0], c[1], c[2]));
}

void print3ds_obj_node_reader (FILE* f, int ind, long p) {
	m3d_fread_name(f);
	M3D_REPORT_CHUNK_DATA(("Object name: \"%s\"", name_buffer));
    chunk_reader(f, ind, p);
}

void print3ds_verts_reader (FILE* f, int ind, long p) {
    word nv;
	unsigned i;
    float v[3];

    m3d_fread_ext(nv,f);

    M3D_REPORT_CHUNK_DATA(("Vertices: %d", (int)nv));

    if(nv>MAX_VERTICES) {
		fprintf(stderr, "Object exceeded MAX_VERTICES count.\n");
		return; 
    }

	for(i=0;i<nv;i++) {
        m3d_fread_ext(v,f);
        M3D_REPORT_CHUNK_DATA(("X: %f, Y: %f, Z: %f", v[0], v[1], v[2]));
    }
	
}

void print3ds_faces_reader (FILE* f, int ind, long p) {
    word nv;
    word face[4];

    m3d_fread_ext(nv,f);
    M3D_REPORT_CHUNK_DATA(("Faces: %d", (int)nv));

    while (nv-- > 0) {
        m3d_fread_ext(face,f);
        M3D_REPORT_CHUNK_DATA(("A: %d, B: %d, C: %d, Flags 0x%X", face[0], face[1], face[2], face[3]));
    }
	
    chunk_reader(f, ind, p);
}

void print3ds_facemat_reader (FILE* f, int ind, long p) {
	word nf, face;
	unsigned i;

	m3d_fread_name(f);
	M3D_REPORT_CHUNK_DATA(("Material name for faces: \"%s\"", name_buffer));

    m3d_fread_ext(nf,f);
    M3D_REPORT_CHUNK_DATA(("Faces with this material: %d", nf));
	
    for(i=0;i<nf;++i) {
        m3d_fread_ext(face,f);
        M3D_REPORT_CHUNK_DATA(("Face: %d", face));
    }
}

void print3ds_maplist_reader (FILE* f, int ind, long p) {
    word nv;
    float uv[2];
	unsigned i;

    m3d_fread_ext(nv,f);

	for(i=0;i<nv;++i) {
        m3d_fread_ext(uv,f);
        M3D_REPORT_CHUNK_DATA(("U: %f, V: %f", uv[0], uv[1]));
    }
	
}

void print3ds_smoothing_groups_reader (FILE* f, int ind, long p) {
    dword s;
    /*int i;*/

    while (ftell(f) < p) {
        m3d_fread_ext(s,f);
        M3D_REPORT_CHUNK_DATA(("Smoothing groups: %lX",s));
        /*for (i = 0; i < 32; i++)
            if (s & (1 << i))
                printf("%d, ", i + 1);
        if(printf_3ds) printf("\n");*/
    }
}

void print3ds_tr_matrix_reader(FILE* f, int ind, long p) {
    float rot[9];
    float trans[3];
    m3d_fread_ext(rot,f);
    M3D_REPORT_CHUNK_DATA(("Rotation matrix:"));
    M3D_REPORT_CHUNK_DATA(("\t%f, %f, %f", rot[0], rot[1], rot[2]));
    M3D_REPORT_CHUNK_DATA(("\t%f, %f, %f", rot[3], rot[4], rot[5]));
    M3D_REPORT_CHUNK_DATA(("\t%f, %f, %f", rot[6], rot[7], rot[8]));
    m3d_fread_ext(trans,f);
    M3D_REPORT_CHUNK_DATA(("Translation matrix: %f, %f, %f", trans[0], trans[1], trans[2]));

}

void print3ds_light_reader(FILE* f, int ind, long p) {
    float l[3];
    m3d_fread_ext(l,f)
    M3D_REPORT_CHUNK_DATA(("X: %f, Y: %f, Z: %f", l[0], l[1], l[2]));

    chunk_reader(f, ind, p);
}

void print3ds_spotlight_reader(FILE* f, int ind, long p) {
    float sl[5];
    m3d_fread_ext(sl,f);
	M3D_REPORT_CHUNK_DATA(("Target X: %f, Y: %f, Z: %f; Hotspot %f, Falloff %f", sl[0], sl[1], sl[2], sl[3], sl[4]));
}
 
void print3ds_camera_reader(FILE* f, int ind, long p) {
    float c[8];
    m3d_fread_ext(c,f);
    M3D_REPORT_CHUNK_DATA(("Position: X: %f, Y: %f, Z: %f", c[0], c[1], c[2]));
    M3D_REPORT_CHUNK_DATA(("Target: X: %f, Y: %f, Z: %f", c[3], c[4], c[5]));
    M3D_REPORT_CHUNK_DATA(("Bank: %f, Lens: %f", c[6], c[7]));
}

void print3ds_mat_name_reader (FILE* f, int ind, long p) {
	m3d_fread_name(f);
	M3D_REPORT_CHUNK_DATA(("Material name \"%s\"", name_buffer));
}

void print3ds_map_file_reader (FILE* f, int ind, long p) {
	m3d_fread_name(f);
    M3D_REPORT_CHUNK_DATA(("Map filename \"%s\"", name_buffer));
}

void print3ds_frame_range_reader(FILE* f, int ind, long p) {
    dword frames[2];
    m3d_fread_ext(frames,f);
    M3D_REPORT_CHUNK_DATA(("Start: %ld, End: %ld", frames[0], frames[1]));
}



void print3ds_m3d_version_reader(FILE*  f, int ind, long p) {
	word version;
	m3d_fread_ext(version,f);
	M3D_REPORT_CHUNK_DATA(("= %d",version));
}

void print3ds_mesh_version_reader(FILE*  f, int ind, long p) {
	word version;
	m3d_fread_ext(version,f);
	M3D_REPORT_CHUNK_DATA(("= %d",version));
}

void print3ds_master_scale_reader(FILE*  f, int ind, long p) {
	float scale;
	m3d_fread_ext(scale,f);
	M3D_REPORT_CHUNK_DATA(("= %f",scale));

}

void print3ds_kfcurtime_reader(FILE*  f, int ind, long p) {
	word curframe;
	m3d_fread_ext(curframe,f);
	M3D_REPORT_CHUNK_DATA(("= %d",curframe));
}

void print3ds_node_id_reader(FILE*  f, int ind, long p) {
	short id;
	m3d_fread_ext(id,f);
	M3D_REPORT_CHUNK_DATA(("= %d",id));
}


void print3ds_node_hdr_reader(FILE*  f, int ind, long p) {
	word flags[2];
	word heirarchy;

	m3d_fread_name(f);
	m3d_fread_ext(flags,f);
	m3d_fread_ext(heirarchy,f);
	M3D_REPORT_CHUNK_DATA(("Name = %s, flags=(0x%X,0x%X), heirarchy=%d",name_buffer,flags[0],flags[1],heirarchy));
}

void print3ds_pivot_reader(FILE*  f, int ind, long p) {
	float pivot_x, pivot_y, pivot_z; 
	m3d_fread_ext(pivot_x,f);
	m3d_fread_ext(pivot_y,f);
	m3d_fread_ext(pivot_z,f);
	M3D_REPORT_CHUNK_DATA(("x,y,z = %f,%f,%f",pivot_x,pivot_y,pivot_z));
}

void print3ds_pos_track_tag_reader(FILE*  f, int ind, long p) {
	struct pos_track_tag_struct {
		word framenum;
		dword unknown;
		float pos_x, pos_y, pos_z; 
	}*  pos;
	word flags;
	word unknown[4];
	word keys;
	word unknown2;
	word i;
	
	m3d_fread_ext(flags,f);
	m3d_fread_ext(unknown,f);
	m3d_fread_ext(keys,f);
	m3d_fread_ext(unknown2,f);

	pos = malloc_(struct pos_track_tag_struct, keys);
	assert(pos != NULL);

	/*
	 * DAMN PADDING MAKES NEXT LINE NOT WORK:
	 * if (keys != fread(pos, sizeof(struct pos_track_tag_struct), keys, f)) { m3d_report_error(p); return; }
	 */
	
	for (i=0;i<keys;++i) {
		m3d_fread_ext(pos[i].framenum, f);
		m3d_fread_ext(pos[i].unknown, f);
		m3d_fread_ext(pos[i].pos_x, f);
		m3d_fread_ext(pos[i].pos_y, f);
		m3d_fread_ext(pos[i].pos_z, f);
	}
		

	M3D_REPORT_CHUNK_DATA(("flags=%d, unknown=(0x%X,0x%X,0x%X,0x%X), keys=%d, unknown2=%d",flags,unknown[0],unknown[1],unknown[2],unknown[3],keys,unknown2));
	for (i=0;i<keys;++i) M3D_REPORT_CHUNK_DATA(("%d)\tframenum=%d, unknown=0x%X, pos=(%f,%f,%f)",
				i,pos[i].framenum,(unsigned int)pos[i].unknown,pos[i].pos_x,pos[i].pos_y,pos[i].pos_z));

	free(pos);
}


void print3ds_rot_track_tag_reader(FILE*  f, int ind, long p) {
	struct rot_track_tag_struct {
		word framenum;
		dword unknown;
		float rotation_rad;
		float axis_x, axis_y, axis_z; 
	}*  rot;
	word flags;
	word unknown[4];
	word keys;
	word unknown2;
	word i;
	
	m3d_fread_ext(flags,f);
	m3d_fread_ext(unknown,f);
	m3d_fread_ext(keys,f);
	m3d_fread_ext(unknown2,f);

	rot = malloc_(struct rot_track_tag_struct, keys);
	assert(rot != NULL);
	
	for (i=0;i<keys;++i) {
		m3d_fread_ext(rot[i].framenum, f);
		m3d_fread_ext(rot[i].unknown, f);
		m3d_fread_ext(rot[i].rotation_rad, f);
		m3d_fread_ext(rot[i].axis_x, f);
		m3d_fread_ext(rot[i].axis_y, f);
		m3d_fread_ext(rot[i].axis_z, f);
	}
		

	M3D_REPORT_CHUNK_DATA(("flags=%d, unknown=(0x%X,0x%X,0x%X,0x%X), keys=%d, unknown2=%d",flags,unknown[0],unknown[1],unknown[2],unknown[3],keys,unknown2));
	for (i=0;i<keys;++i) M3D_REPORT_CHUNK_DATA(("%d)\tframenum=%d, unknown=0x%X, rotation=%f, axis=(%f,%f,%f)",
				i,rot[i].framenum,(unsigned int)rot[i].unknown,rot[i].rotation_rad, rot[i].axis_x,rot[i].axis_y,rot[i].axis_z));

	free(rot);
}


void print3ds_scl_track_tag_reader(FILE*  f, int ind, long p) {
	struct pos_track_tag_struct {
		word framenum;
		dword unknown;
		float scale_x, scale_y, scale_z; 
	}*  scl;
	word flags;
	word unknown[4];
	word keys;
	word unknown2;
	word i;
	
	m3d_fread_ext(flags,f);
	m3d_fread_ext(unknown,f);
	m3d_fread_ext(keys,f);
	m3d_fread_ext(unknown2,f);

	scl = malloc_(struct pos_track_tag_struct, keys);
	assert(scl != NULL);	

	for (i=0;i<keys;++i) {
		m3d_fread_ext(scl[i].framenum, f);
		m3d_fread_ext(scl[i].unknown, f);
		m3d_fread_ext(scl[i].scale_x, f);
		m3d_fread_ext(scl[i].scale_y, f);
		m3d_fread_ext(scl[i].scale_z, f);
	}
		

	M3D_REPORT_CHUNK_DATA(("flags=%d, unknown=(0x%X,0x%X,0x%X,0x%X), keys=%d, unknown2=%d",flags,unknown[0],unknown[1],unknown[2],unknown[3],keys,unknown2));
	for (i=0;i<keys;++i) M3D_REPORT_CHUNK_DATA(("%d)\tframenum=%d, unknown=0x%X, scale=(%f,%f,%f)",
				i,scl[i].framenum,(unsigned int)scl[i].unknown,scl[i].scale_x,scl[i].scale_y,scl[i].scale_z));

	free(scl);
}

void print3ds_kfheader_reader(FILE*  f, int ind, long p) {
	word revision; 
	word animlen; 

	m3d_fread_ext(revision,f);
	m3d_fread_name(f);
	m3d_fread_ext(animlen,f);
	M3D_REPORT_CHUNK_DATA(("Name = %s, revision=%d, animlen=%d",name_buffer,revision,animlen));
}


void print3ds_instancename_reader(FILE*  f, int ind, long p) {
	m3d_fread_name(f);
	M3D_REPORT_CHUNK_DATA(("Name = %s",name_buffer));
}

/* Currently NULL means the chunk has no content, and the program should proceed to its child-chunks. */
/* SKIP_CHUNK means the chunk may or may not contain data, but the program should disregard it anyway. */
/* Also, the main chunk_reader cannot be provided. */
/* I know it sucks. I should change it. */
/*#ifndef JUSTATRY*/
/*#define JUSTATRY*/
/*#endif*/
chunk_handler print3ds_handlers_list[] = {

	{CHUNK_M3D_VERSION,	print3ds_m3d_version_reader},
	{CHUNK_RGBF,	print3ds_RGBF_reader},
	{CHUNK_RGBB,	print3ds_RGBB_reader},
	{CHUNK_MAIN,	NULL},
	{CHUNK_OBJMESH,	NULL},
	{CHUNK_MESH_VERSION,	print3ds_mesh_version_reader},
	{CHUNK_MASTER_SCALE,	print3ds_master_scale_reader},
	{CHUNK_BKGCOLOR,	NULL},
	{CHUNK_AMBCOLOR,	NULL},
	{CHUNK_OBJBLOCK,	print3ds_obj_node_reader},
	{CHUNK_TRIMESH,	NULL},
	{CHUNK_VERTLIST,	print3ds_verts_reader},
	{CHUNK_FACELIST,	print3ds_faces_reader},
	{CHUNK_FACEMAT,	print3ds_facemat_reader},
	{CHUNK_MAPLIST,	print3ds_maplist_reader},
	{CHUNK_SMOOLIST,	print3ds_smoothing_groups_reader},
	{CHUNK_TRMATRIX,	print3ds_tr_matrix_reader},
	{CHUNK_LIGHT,	print3ds_light_reader},
	{CHUNK_SPOTLIGHT,	print3ds_spotlight_reader},
	{CHUNK_CAMERA,	print3ds_camera_reader},
	{CHUNK_MATERIAL,	NULL},
	{CHUNK_MATNAME,	print3ds_mat_name_reader},
	{CHUNK_AMBIENT,	NULL},
	{CHUNK_DIFFUSE,	NULL},
	{CHUNK_SPECULAR,	NULL},
	{CHUNK_TEXTURE,	NULL},
	{CHUNK_BUMPMAP,	NULL},
	{CHUNK_MAPFILE,	print3ds_map_file_reader},
	{CHUNK_KEYFRAMER,	NULL},
	{CHUNK_KFHDR,	print3ds_kfheader_reader},
	{CHUNK_FRAMES,	print3ds_frame_range_reader},
	{CHUNK_KFCURTIME,	print3ds_kfcurtime_reader},
	{CHUNK_OBJECT_NODE_TAG,	NULL},
	{CHUNK_CAMERA_NODE_TAG,	NULL},
	{CHUNK_TARGET_NODE_TAG,	NULL},
	{CHUNK_NODE_ID,	print3ds_node_id_reader},
	{CHUNK_NODE_HDR,	print3ds_node_hdr_reader},
	{CHUNK_PIVOT,	print3ds_pivot_reader},
	{CHUNK_INSTANCE_NAME,	print3ds_instancename_reader},
	{CHUNK_POS_TRACK_TAG,	print3ds_pos_track_tag_reader},
	{CHUNK_ROT_TRACK_TAG,	print3ds_rot_track_tag_reader},
	{CHUNK_SCL_TRACK_TAG,	print3ds_scl_track_tag_reader},
	{CHUNK_FOV_TRACK_TAG,	SKIP_CHUNK},
	{CHUNK_ROL_TRACK_TAG,	SKIP_CHUNK},
	{CHUNK_BOUNDBOX,	SKIP_CHUNK},
	{CHUNK_NULL, NULL},
};


/*------------------==[ End of print 3DS handlers ]==------------------------*/


