#include <vector>
#include <GLee/GLee.h>
#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#include "Engine.h"
#include "CustomSceneNodes.h"
#include "ShaderManager.h"

namespace GameEngine {

	Engine::Engine() :
	  rootSN_(new SceneNode),
	  terrain_(new Terrain(1, 1)),
	  models_(),
	  materials_(),
	  picker_(new Picker),
	  fps_(new FPS),
	  drawSoftShadows_(true) {
		Ptr<SceneNode> tsn = new TerrainSceneNode(terrain_);
		terrain_->sceneNodeIs(tsn);
		rootSN_->childNew(tsn);
	}

	void Engine::initAll() {
		initDisplay();
		initGLSettings();
		initMyStuff();
		initLighting();
		initMatrixes();
		initTextures();
		initSoftShadows();
	}

	void Engine::initGLSettings() {
		glClearDepth(1.0f);
		glEnable(GL_DEPTH_TEST);
		glDepthFunc(GL_LEQUAL);
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

		glShadeModel(GL_SMOOTH);

		glCullFace(GL_BACK);

		glClearColor(0.0, 0.0, 0.0, 0.0);
	}

	void Engine::resize(int w, int h) {
		SDL_SetVideoMode(w, h, WINDOW_BPP, SDL_OPENGL);
		glViewport(0, 0, (GLsizei)w, (GLsizei)h);
		initMatrixes();
	}

	void Engine::initDisplay() {
		SDL_Init(SDL_INIT_EVERYTHING);
		putenv("SDL_VIDEO_WINDOW_POS=center");
		putenv("SDL_VIDEO_CENTERED=1");
		SDL_SetVideoMode(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_BPP, SDL_OPENGL);
		SDL_WM_SetCaption(WINDOW_TITLE, NULL);
		glViewport(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
	}

	void Engine::initLighting() {
		GLfloat light0_pos[] = {LIGHT_X, LIGHT_Y, LIGHT_Z, 1.0};
		//GLfloat light0_color[] = {1.0, 1.0, 1.0, 1.0};
		//GLfloat ambient_light[] = {0.8f, 0.8f, 0.8f, 1.0};

		float ambience[4] = {0.3f, 0.3f, 0.3f, 1.0};		// The color of the light in the world
		float diffuse[4] = {0.8f, 0.8f, 0.8f, 1.0};			// The color of the positioned light

		glLightfv( GL_LIGHT0, GL_AMBIENT,  ambience );
		glLightfv( GL_LIGHT0, GL_DIFFUSE,  diffuse );

		/* Set up some light sources */
		glLightfv(GL_LIGHT0, GL_POSITION, light0_pos);
		//glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_color);
		//glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient_light);

		glEnable(GL_LIGHT0);
		glEnable(GL_LIGHTING);

		// TODO: temp only
		glEnable(GL_COLOR_MATERIAL);
		glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
	}

	void Engine::initTextures() {
		/* Due to textures loaded in mirrored and flipped, need to correct:
		 * http://steinsoft.net/index.php?site=Programming/Code%20Snippets/SDL/no1
		 */
		glMatrixMode(GL_TEXTURE);
		glRotatef(180.0f,0.0f,0.0f,1.0f);
		glScalef(-1.0f,1.0f,1.0f);
		glMatrixMode(GL_MODELVIEW);

		vector<Ptr<Model> >::iterator modelIter = models_.begin();
		while (modelIter != models_.end()) {
			Ptr<Model> model = *modelIter;

			vector<Ptr<Material> > modelMatVec = model->loadMaterials();
			vector<Ptr<Material> >::iterator matIter = modelMatVec.begin();
			while (matIter != modelMatVec.end()) {
				// TODO: only add unique textures
				materials_.push_back(*matIter);
				matIter++;
			}

			modelIter++;
		}
	}

	void Engine::initMatrixes() {
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(FOV_ANGLE, WINDOW_WIDTH/WINDOW_HEIGHT, CLIPPLANE_NEAR, 
			CLIPPLANE_FAR);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		gluLookAt(
			VIEWER_X, VIEWER_Y, VIEWER_Z,
			CENTER_X, CENTER_Y, CENTER_Z, 
			UP_X, UP_Y, UP_Z);
	}

	void Engine::initMyStuff() {
	}

	void Engine::exitEngine() {
		SDL_Quit();
		exit(0);
	}

	unsigned int Engine::processPick(bool terrainOnly, unsigned int x, unsigned int y) {
		Ptr<Timer> t = new Timer;
		t->start();

		picker_->startPicking(x, y);

		if (terrainOnly) {
			terrain_->renderLevelIs(Terrain::RLSimple);
			glPushMatrix();
			rootSN_->applyTransform();
			terrain_->sceneNode()->render();
			glPopMatrix();
			terrain_->renderLevelIs(Terrain::RLFull);
		} else {
			rootSN_->render();
		}

		picker_->stopPicking();

		return picker_->pickedId();
	}

	void Engine::onPick(unsigned int pickedId) { }

	void Engine::handleEvents() {
		SDL_Event e;
		while (SDL_PollEvent(&e)) {
			handleEvent(e);
		}
	}

	void Engine::handleEvent(SDL_Event &e) {
		switch (e.type) {
			case SDL_QUIT:
				exitEngine();
			case SDL_VIDEORESIZE:
				resize(e.resize.w, e.resize.h);
			case SDL_MOUSEMOTION:
				onMouseMotion(e.motion.x, e.motion.y);
				break;
			case SDL_MOUSEBUTTONDOWN:
				onMouseBtnDown(e.button.button, e.button.x, e.button.y);
				break;
			case SDL_MOUSEBUTTONUP:
				onMouseBtnUp(e.button.button, e.button.x, e.button.y);
				break;
			case SDL_KEYDOWN:
				onKeyDown(e.key.keysym.sym, e.key.keysym.mod);
				break;
			case SDL_KEYUP:
				onKeyUp(e.key.keysym.sym, e.key.keysym.mod);
			default:
				break;
		}
	}

	void Engine::onMouseMotion(unsigned int x, unsigned int y) {
		bool consumed = hud_->onMouseMotion(x, WINDOW_HEIGHT - y);
		if (!consumed) {
		}
	}

	void Engine::onMouseBtnDown(unsigned int button, unsigned int x, unsigned int y) {
		bool consumed = hud_->onMouseDown(button, x, WINDOW_HEIGHT - y);
		if (!consumed) {
			switch (button) {
				case SDL_BUTTON_LEFT:
					onPick(processPick(false, x, y));
					break;
				case SDL_BUTTON_RIGHT:
					break;
				case SDL_BUTTON_WHEELDOWN:
					rootSN_->translate(0, 0, -1);
					break;
				case SDL_BUTTON_WHEELUP:
					rootSN_->translate(0, 0, 1);
					break;
				default:
					break;
			}
		}
	}

	void Engine::onMouseBtnUp(unsigned int button, unsigned int x, unsigned int y) {
	}

	void Engine::onKeyDown(SDLKey key, SDLMod modifer) {
		// escape key
		switch (key) {
			case SDLK_ESCAPE:
				exitEngine();
				break;
			case SDLK_UP:
				rootSN_->rotate(5, -1, 0, 0);
				break;
			case SDLK_DOWN:
				rootSN_->rotate(5, 1, 0, 0);
				break;
			case SDLK_w: {
				Ptr<Vec3> lookVec = new Vec3(VIEWER_X, VIEWER_Y, VIEWER_Z);
				lookVec->normalize();
				lookVec = rootSN_->worldTransform()->mult(lookVec);
				rootSN_->translate(lookVec->x, lookVec->y, lookVec->z);
				break;
			}
			case SDLK_b:
				ShaderManager::Instance()->shader(ShaderManager::BUMPMAP_SHADER)->modeInc();
				break;
			default:
				break;
		}
	}

	void Engine::onKeyUp(SDLKey key, SDLMod modifer) {
	}

	void Engine::gameLoop() {
		unsigned int elapsedMillis;
		fps_->start();

		while (true) {
			elapsedMillis = fps_->elapsedMillis();
			fps_->update();

			handleEvents();
			update(elapsedMillis);

			draw();
			
			glFlush();
			SDL_GL_SwapBuffers();

			fps_->delayToRegulateFPS(FRAMES_PER_SECOND);
		}
	}

	void Engine::update(float elapsedMillis) {
	}

	void Engine::draw() {
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		if (drawSoftShadows_) {
			renderWithShadows();
		} else {
			rootSN_->render();
		}

		hud_->draw();
	}

	void Engine::run(int argc, char **argv) {
		gameLoop();
	}

	//
	// Soft shadows
	//

	void Engine::initSoftShadows() {
		//
		// INIT PART
		//
		cameraPosition_ = new Vec3(VIEWER_X, VIEWER_Y, VIEWER_Z);
		lightPosition_ = new Vec3(LIGHT_X, LIGHT_Y, LIGHT_Z);

		 //Size of shadow map
		shadowMapSize_ = 512;

		//Check for necessary extensions
		if(!GLEE_ARB_depth_texture || !GLEE_ARB_shadow)
		{
			cout << "I require ARB_depth_texture and ARB_shadow extensionsn" << endl;
			return;
		}

		//Shading states
		glShadeModel(GL_SMOOTH);
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

		//Depth states
		glClearDepth(1.0f);
		glDepthFunc(GL_LEQUAL);
		glEnable(GL_DEPTH_TEST);

		glEnable(GL_CULL_FACE);

		//We use glScale when drawing the scene
		glEnable(GL_NORMALIZE);

		//Create the shadow map texture
		glGenTextures(1, &shadowMapTexture_);
		glBindTexture(GL_TEXTURE_2D, shadowMapTexture_);
		glTexImage2D( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, shadowMapSize_, shadowMapSize_, 0,
			GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, NULL);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

		//Use the color as the ambient and diffuse material
		glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);

		//White specular material color, shininess 16
		float white[4] = { 1, 1, 1, 1 };
		glMaterialfv(GL_FRONT, GL_SPECULAR, white);
		glMaterialf(GL_FRONT, GL_SHININESS, 16.0f);

		//Calculate & save matrices
		float t[16];

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();

		glLoadIdentity();
		gluPerspective(FOV_ANGLE, WINDOW_WIDTH/WINDOW_HEIGHT, CLIPPLANE_NEAR, CLIPPLANE_FAR);
		glGetFloatv(GL_MODELVIEW_MATRIX, t);
		cameraProjectionMatrix_ = new Matrix(t);

		glLoadIdentity();
		gluLookAt(VIEWER_X, VIEWER_Y, VIEWER_Z,
				  CENTER_X, CENTER_Y, CENTER_Z, 
				  UP_X, UP_Y, UP_Z);
		glGetFloatv(GL_MODELVIEW_MATRIX, t);
		cameraViewMatrix_ = new Matrix(t);

		glLoadIdentity();
		gluPerspective(45.0f, 1.0f, 1, 100.0f);
		glGetFloatv(GL_MODELVIEW_MATRIX, t);
		lightProjectionMatrix_ = new Matrix(t);

		glLoadIdentity();
		gluLookAt(LIGHT_X, LIGHT_Y, LIGHT_Z,
				  CENTER_X, CENTER_Y, CENTER_Z, 
				  UP_X, UP_Y, UP_Z);
		glGetFloatv(GL_MODELVIEW_MATRIX, t);
		lightViewMatrix_ = new Matrix(t);

		glPopMatrix();
	}

	void Engine::renderWithShadows() {
		//
		// DRAW 1ST PASS
		//

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();

		//First pass - from light's point of view
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		rootSN_->render();
/*
		float t[16];
		glMatrixMode(GL_PROJECTION);
		lightProjectionMatrix_->putValues(t);
		glLoadMatrixf(t);

		glMatrixMode(GL_MODELVIEW);
		lightViewMatrix_->putValues(t);
		glLoadMatrixf(t);

		//Use viewport the same size as the shadow map
		glViewport(0, 0, shadowMapSize_, shadowMapSize_);

		//Draw back faces into the shadow map
		glCullFace(GL_FRONT);

		//Disable color writes, and use flat shading for speed
		glShadeModel(GL_FLAT);
		glColorMask(0, 0, 0, 0);

		//Draw the scene
		//rootSN_->render();
		testScene();

		//Read the depth buffer into the shadow map texture
		glBindTexture(GL_TEXTURE_2D, shadowMapTexture_);
		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, shadowMapSize_, shadowMapSize_);

		//restore states
		glCullFace(GL_BACK);
		glShadeModel(GL_SMOOTH);
		glColorMask(1, 1, 1, 1);
*/
/*
		glViewport(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);

			glPushMatrix();
			glScalef(10, 10, 10);

			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE);

			glBindTexture(GL_TEXTURE_2D, shadowMapTexture_);
			glEnable(GL_TEXTURE_2D);

			glColor4f(1, 1, 1, 1);

			glBegin(GL_QUADS);
			glNormal3f(0, 0, 1);
			glTexCoord2f(0, 0);
			glVertex3f(-0.5, -0.5, 0);
			glTexCoord2f(1, 0);
			glVertex3f(0.5, -0.5, 0);
			glTexCoord2f(1, 1);
			glVertex3f(0.5, 0.5, 0);
			glTexCoord2f(0, 1);
			glVertex3f(-0.5, 0.5, 0);
			glEnd();

			glDisable(GL_TEXTURE_2D);

			glDisable(GL_BLEND);

			glPopMatrix();
*/
		//
		// DRAW 2ND PASS
		//
/*
		//2nd pass - Draw from camera's point of view
		glClear(GL_DEPTH_BUFFER_BIT);
		glMatrixMode(GL_PROJECTION);
		cameraProjectionMatrix_->putValues(t);
		glLoadMatrixf(t);

		glMatrixMode(GL_MODELVIEW);
		cameraViewMatrix_->putValues(t);
		glLoadMatrixf(t);

		glViewport(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);

		//Use dim light to represent shadowed areas
		float lightPos[4] = { lightPosition_->x, lightPosition_->y, lightPosition_->z, 1 };
		float black[4] = { 0, 0, 0, 0 };
		float whiteDim[4] = { 0.2, 0.2, 0.2, 0.2 };
		float white[4] = { 1, 1, 1, 1 };
		glLightfv(GL_LIGHT1, GL_POSITION, lightPos);
		glLightfv(GL_LIGHT1, GL_AMBIENT, whiteDim);
		glLightfv(GL_LIGHT1, GL_DIFFUSE, whiteDim);
		glLightfv(GL_LIGHT1, GL_SPECULAR, black);
		glEnable(GL_LIGHT1);
		glEnable(GL_LIGHTING);

		//Draw the scene
		//rootSN_->render();
		testScene();
*/
		//
		// DRAW 3RD PASS
		//
/*
		//3rd pass
		//Draw with bright light
		glLightfv(GL_LIGHT1, GL_DIFFUSE, white);
		glLightfv(GL_LIGHT1, GL_SPECULAR, white);

		//Calculate texture matrix for projection
		//This matrix takes us from eye space to the light's clip space
		//It is postmultiplied by the inverse of the current view matrix when specifying texgen
		float biasMatrix[16] = {0.5f, 0.0f, 0.0f, 0.0f,
								0.0f, 0.5f, 0.0f, 0.0f,
								0.0f, 0.0f, 0.5f, 0.0f,
								0.5f, 0.5f, 0.5f, 1.0f}; //bias from [-1, 1] to [0, 1]
		Ptr<Matrix> biasM = new Matrix(biasMatrix);
		Ptr<Matrix> textureM = biasM->matrixMult(lightProjectionMatrix_)->matrixMult(lightViewMatrix_);
		float row0[4], row1[4], row2[4], row3[4];
		textureM->row(0, row0);
		textureM->row(1, row1);
		textureM->row(2, row2);
		textureM->row(3, row3);

		//Set up texture coordinate generation.
		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGenfv(GL_S, GL_EYE_PLANE, row0);
		glEnable(GL_TEXTURE_GEN_S);

		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGenfv(GL_T, GL_EYE_PLANE, row1);
		glEnable(GL_TEXTURE_GEN_T);

		glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGenfv(GL_R, GL_EYE_PLANE, row2);
		glEnable(GL_TEXTURE_GEN_R);

		glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGenfv(GL_Q, GL_EYE_PLANE, row3);
		glEnable(GL_TEXTURE_GEN_Q);

		//Bind & enable shadow map texture
		glBindTexture(GL_TEXTURE_2D, shadowMapTexture_);
		glEnable(GL_TEXTURE_2D);

		//Enable shadow comparison
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE);

		//Shadow comparison should be true (ie not in shadow) if r<=texture
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);

		//Shadow comparison should generate an INTENSITY result
		glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE_ARB, GL_INTENSITY);

		//Set alpha test to discard false comparisons
		glAlphaFunc(GL_GEQUAL, 0.1f);
		glEnable(GL_ALPHA_TEST);

		//Draw the scene
		//rootSN_->render();
		testScene();

		//Disable textures and texgen
		glDisable(GL_TEXTURE_2D);

		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		glDisable(GL_TEXTURE_GEN_R);
		glDisable(GL_TEXTURE_GEN_Q);

		//Restore other states
		glDisable(GL_LIGHTING);
		glDisable(GL_ALPHA_TEST);
*/
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
	}

	void Engine::testScene() {
		//Display lists for objects
		static GLuint spheresList=0, torusList=0, baseList=0;

		//Create spheres list if necessary
		if(!spheresList)
		{
			spheresList=glGenLists(1);
			glNewList(spheresList, GL_COMPILE);
			{
				glColor3f(0.0f, 1.0f, 0.0f);
				glPushMatrix();

				glTranslatef(0.45f, 1.0f, 0.45f);
				glutSolidSphere(0.2, 24, 24);

				glTranslatef(-0.9f, 0.0f, 0.0f);
				glutSolidSphere(0.2, 24, 24);

				glTranslatef(0.0f, 0.0f,-0.9f);
				glutSolidSphere(0.2, 24, 24);

				glTranslatef(0.9f, 0.0f, 0.0f);
				glutSolidSphere(0.2, 24, 24);

				glPopMatrix();
			}
			glEndList();
		}

		//Create torus if necessary
		if(!torusList)
		{
			torusList=glGenLists(1);
			glNewList(torusList, GL_COMPILE);
			{
				glColor3f(1.0f, 0.0f, 0.0f);
				glPushMatrix();

				glTranslatef(0.0f, 0.5f, 0.0f);
				glRotatef(90.0f, 1.0f, 0.0f, 0.0f);
				glutSolidTorus(0.2, 0.5, 24, 48);

				glPopMatrix();
			}
			glEndList();
		}

		//Create base if necessary
		if(!baseList)
		{
			baseList=glGenLists(1);
			glNewList(baseList, GL_COMPILE);
			{
				glColor3f(0.0f, 0.0f, 1.0f);
				glPushMatrix();

				glScalef(1.0f, 0.05f, 1.0f);
				glutSolidCube(3.0f);

				glPopMatrix();
			}
			glEndList();
		}


		//Draw objects
		glPushMatrix();
		glRotatef(45, 1, 0, 0);
		glCallList(baseList);
		glCallList(torusList);

		glPushMatrix();
		static float angle = 0;
		angle += 0.01;
		//glRotatef(angle, 0.0f, 1.0f, 0.0f);
		glScalef(angle, angle, angle);
		glCallList(spheresList);
		glPopMatrix();
		glPopMatrix();
	}
}