#include <SDL/SDL_opengl.h>
#include <math.h>
#include <string>
#include <iostream>
#include <sstream>

#include "Geometry.h"
#include "Drawable.h"
#include "Ptr.h"

namespace GameEngine {

	const float Math::PI = 3.14159265358979;

	float Math::radians(float degrees) {
		return degrees * Math::PI / 180.0;
	}

	float Math::degrees(float radians) {
		return radians * 180 / Math::PI;
	}

	float Math::det(float a, float b, float c, float d) {
		return a * d - b * c;
	}

	Ptr<Vec3> Math::computeNormal(Ptr<Vertex3> v1, Ptr<Vertex3> v2, Ptr<Vertex3> v3) {
		Ptr<Vec3> vec1 = v2->sub(v1);
		Ptr<Vec3> vec2 = v3->sub(v1);
		Ptr<Vec3> normal = vec1->cross(vec2);
		normal->normalize();
		return normal;
	}

	float Vec3::mag() {
		return sqrtf(x*x + y*y + z*z);
	}

	void Vec3::normalize() {
		float scale = mag();
		x /= scale;
		y /= scale;
		z /= scale;
	}

	Ptr<Vec3> Vec3::cross(const Ptr<Vec3> v) {
		Ptr<Vec3> cross = new Vec3;
		cross->x = Math::det(y, v->y, z, v->z);
		cross->y = -1 * Math::det(x, v->x, z, v->z);
		cross->z = Math::det(x, v->x, y, v->y);
		return cross;
	}

	Ptr<Vec3> Vec3::mult(float v) {
		Ptr<Vec3> mult = new Vec3;
		mult->x = x * v;
		mult->y = y * v;
		mult->z = z * v;
		return mult;
	}

	Ptr<Vec3> Vec3::mult(const Ptr<Vec3> v) {
		Ptr<Vec3> mult = new Vec3;
		mult->x = x * v->x;
		mult->y = y * v->y;
		mult->z = z * v->z;
		return mult;
	}

	Ptr<Vec3> Vec3::plus(const Ptr<Vec3> v) {
		Ptr<Vec3> plus = new Vec3;
		plus->x = x + v->x;
		plus->y = y + v->y;
		plus->z = z + v->z;
		return plus;
	}

	Ptr<Vec3> Vec3::sub(const Ptr<Vec3> v) {
		Ptr<Vec3> sub = new Vec3;
		sub->x = x + v->x;
		sub->y = y + v->y;
		sub->z = z + v->z;
		return sub;
	}

	string Vec3::toString() {
		stringstream ss;
		ss << "[" << x << ", " << y << ", " << z << "]";
		return ss.str();
	}

	Ptr<Vec3> Vertex3::sub(const Ptr<Vertex3> v) {
		Ptr<Vec3> subv = new Vec3;
		subv->x = x - v->x;
		subv->y = y - v->y;
		subv->z = z - v->z;
		return subv;
	}

	string Vertex3::toString() {
		stringstream ss;
		ss << "[" << x << ", " << y << ", " << z << "]";
		return ss.str();
	}

	string Mapping::toString() {
		stringstream ss;
		ss << "[" << u << ", " << v << "]";
		return ss.str();
	}

	string Face::toString() {
		stringstream ss;
		ss << "[" << a << ", " << b << ", " << c << "] flags=[" << flags;
		ss << "] matName=[" << materialName << "]";
		return ss.str();
	}

	string RGBColor::toString() {
		stringstream ss;
		ss << "[" << r << ", " << g << ", " << b << "]";
		return ss.str();
	}

	void Matrix::apply() {
		glMultMatrixf(values_);
	}

	void Matrix::row(unsigned int i, float* result) {
		result[0] = values_[i * 4 + 0];
		result[1] = values_[i * 4 + 1];
		result[2] = values_[i * 4 + 2];
		result[3] = values_[i * 4 + 3];
	}

	void Matrix::translate(float x, float y, float z) {
		Ptr<Matrix> m = new Matrix;

		// change the last column
		m->value(12) = x;
		m->value(13) = y;
		m->value(14) = z;

		m = this->matrixMult(m);
		m->putValues(values_);
	}

	void Matrix::scale(float x, float y, float z) {
		Ptr<Matrix> m = new Matrix;

		// change the diagonal
		m->value(0) = x;
		m->value(5) = y;
		m->value(10) = z;

		m = this->matrixMult(m);
		m->putValues(values_);
	}

	// angle expects degrees
	void Matrix::rotate(float angle, float x, float y, float z) {
		// normalize the direction vector
		Ptr<Vec3> vnorm = new Vec3;
		vnorm->x = x;
		vnorm->y = y;
		vnorm->z = z;
		vnorm->normalize();
		x = vnorm->x;
		y = vnorm->y;
		z = vnorm->z;

		// convert angle to radians
		float radangle = Math::radians(angle);

		Ptr<Matrix> m = new Matrix;
		
		/* Use equation from gamedev.net for rotating about arbitrary axis:
		 * http://www.gamedev.net/reference/articles/article1199.asp
		 */
		float c = cosf(radangle);
		float s = sinf(radangle);
		float t = 1 - cosf(radangle);

		m->value(0) = t * x * x + c;
		m->value(1) = t * x * y + s * z;
		m->value(2) = t * x * z - s * y;
		m->value(3) = 0;

		m->value(4) = t * x * y - s * z;
		m->value(5) = t * y * y + c;
		m->value(6) = t * y * z + s * x;
		m->value(7) = 0;

		m->value(8) = t * x * z + s * y;
		m->value(9) = t * y * z - s * x;
		m->value(10) = t * z * z + c;
		m->value(11) = 0;

		m->value(12) = 0;
		m->value(13) = 0;
		m->value(14) = 0;
		m->value(15) = 1;

		m = this->matrixMult(m);
		m->putValues(values_);
	}

	Ptr<Matrix> Matrix::matrixMult(const Ptr<Matrix> other) {
		Ptr<Matrix> m = new Matrix;

		for (int i = 0; i < 4; i++) {
			for (int j = 0; j < 4; j++) {
				m->value(i * 4 + j) = 0;
				for (int k = 0; k < 4; k++) {
					m->value(i * 4 + j) += values_[k * 4 + j] * other->value(i * 4 + k);
				}
			}
		}

		return m;
	}

	Ptr<Vec3> Matrix::mult(Ptr<Vec3> v) {
		// XXX: ick, gross
		Ptr<Matrix> hack = new Matrix;
		hack->value(0) = v->x;
		hack->value(1) = v->y;
		hack->value(2) = v->z;

		hack = matrixMult(hack);

		Ptr<Vec3> ret = new Vec3;
		ret->x = hack->value(0);
		ret->y = hack->value(1);
		ret->z = hack->value(2);

		return ret;
	}
	
	void Matrix::setToIdentity() {
		values_[0] = 1; values_[4] = 0; values_[8] = 0; values_[12] = 0;
		values_[1] = 0; values_[5] = 1; values_[9] = 0; values_[13] = 0;
		values_[2] = 0; values_[6] = 0; values_[10] = 1; values_[14] = 0;
		values_[3] = 0; values_[7] = 0; values_[11] = 0; values_[15] = 1;
	}

	string Matrix::toString() {
		stringstream ss;

		ss << "Matrix:" << endl;
		ss << "[" << values_[0] << ", " << values_[4] << ", " << values_[8] << ", " << values_[12] << "]" << endl;
		ss << "[" << values_[1] << ", " << values_[5] << ", " << values_[9] << ", " << values_[13] << "]" << endl;
		ss << "[" << values_[2] << ", " << values_[6] << ", " << values_[10] << ", " << values_[14] << "]" << endl;
		ss << "[" << values_[3] << ", " << values_[7] << ", " << values_[11] << ", " << values_[15] << "]" << endl;

		return ss.str();
	}

	void SubMesh::drawMesh() {
		glBegin(GL_TRIANGLES);

		Ptr<Vertex3> v;
		Ptr<Mapping> m;
		for (int i = 0; i < faces.size(); i++) {
			glNormal3f(normals[i]->x, normals[i]->y, normals[i]->z);

			v = vertexes[faces[i]->a];
			if (mappings.size() == vertexes.size()) {
				m = mappings[faces[i]->a];
				glTexCoord2f(m->u, m->v);
			}
			glVertex3f(v->x, v->y, v->z);

			v = vertexes[faces[i]->b];
			if (mappings.size() == vertexes.size()) {
				m = mappings[faces[i]->b];
				glTexCoord2f(m->u, m->v);
			}
			glVertex3f(v->x, v->y, v->z);

			v = vertexes[faces[i]->c];
			if (mappings.size() == vertexes.size()) {
				m = mappings[faces[i]->c];
				glTexCoord2f(m->u, m->v);
			}
			glVertex3f(v->x, v->y, v->z);
		}

		glEnd();
	}

	void SubMesh::draw() {
		// for selecting objects in 3d
		glPushName(id_);

		// draw
		material->bind();
		if (drawId_ > 0) {
			glCallList(drawId_);
		} else {
			drawMesh();
		}
		material->unbind();

		// for selecting objects in 3d
		glPopName();
	}

	void SubMesh::build() {
		if (drawId_ > 0) {
			glDeleteLists(drawId_, 1);
		}
		drawId_ = glGenLists(1);
		glNewList(drawId_, GL_COMPILE);

//		if (texture != null) {
//			this.texture.Bind();
//		} else {
//			Engine.setMaterialProperties(Color.DEFAULT);
//		}

		drawMesh();

//		if (texture != null) {
//			Engine.unbindTexture();
//		}

		glEndList();
	}

	Ptr<Drawable> SubMesh::copy() {
		return new SubMesh;
	}

	string SubMesh::toString() {
		stringstream ss;

		ss << "--- SUBMESH ---" << endl;
		ss << "Name: [" << name() << "]" << endl;
		ss << "ID: [" << id() << "]" << endl;

		ss << "-------------------------------------" << endl;
		ss << "Vertexes:" << endl;
		ss << "-------------------------------------" << endl;
		unsigned int i = 0;
		vector<Ptr<Vertex3> >::iterator vIter = vertexes.begin();
		while (vIter != vertexes.end()) {
			ss << i << ":\t" << (*vIter).ptr() << endl;
			vIter++;
			i++;
		}

		ss << "-------------------------------------" << endl;
		ss << "Mappings:" << endl;
		ss << "-------------------------------------" << endl;
		i = 0;
		vector<Ptr<Mapping> >::iterator mIter = mappings.begin();
		while (mIter != mappings.end()) {
			ss << i << ":\t" << (*mIter).ptr() << endl;
			mIter++;
			i++;
		}

		ss << "-------------------------------------" << endl;
		ss << "Faces:" << endl;
		ss << "-------------------------------------" << endl;
		i = 0;
		vector<Ptr<Face> >::iterator fIter = faces.begin();
		while (fIter != faces.end()) {
			ss << i << ":\t" << (*fIter).ptr() << endl;
			fIter++;
			i++;
		}

		ss << "-------------------------------------" << endl;
		ss << "Normals:" << endl;
		ss << "-------------------------------------" << endl;
		i = 0;
		vector<Ptr<Vec3> >::iterator nIter = normals.begin();
		while (nIter != normals.end()) {
			ss << i << ":\t" << (*nIter).ptr() << endl;
			nIter++;
			i++;
		}

		return ss.str();
	}

	void Mesh::drawMesh() {
		map<string, Ptr<SubMesh> >::iterator subIter = submeshes.begin();
		while (subIter != submeshes.end()) {
			Ptr<SubMesh> submesh = (*subIter).second;
			submesh->draw();
			subIter++;
		}
	}

	void Mesh::draw() {
		// for selecting objects in 3d
		glPushName(id_);

		// draw
		//if (drawId_ > 0) {
		//	glCallList(drawId_);
		//} else {
			drawMesh();
		//}

		// for selecting objects in 3d
		glPopName();
	}

	void Mesh::build() {
		// build submeshes
		map<string, Ptr<SubMesh> >::iterator subIter = submeshes.begin();
		while (subIter != submeshes.end()) {
			Ptr<SubMesh> submesh = (*subIter).second;
			submesh->build();
			subIter++;
		}

		if (drawId_ > 0) {
			glDeleteLists(drawId_, 1);
		}
		drawId_ = glGenLists(1);
		glNewList(drawId_, GL_COMPILE);

//		if (texture != null) {
//			this.texture.Bind();
//		} else {
//			Engine.setMaterialProperties(Color.DEFAULT);
//		}

		drawMesh();

//		if (texture != null) {
//			Engine.unbindTexture();
//		}

		glEndList();
	}

	Ptr<Drawable> Mesh::copy() {
		return new Mesh;
	}

	string Mesh::toString() {
		stringstream ss;

		ss << "--- MESH ---" << endl;
		ss << "Name: [" << name() << "]" << endl;
		ss << "ID: [" << id() << "]" << endl;

		ss << "-------------------------------------" << endl;
		ss << "SubMeshes:" << endl;
		ss << "-------------------------------------" << endl;
		map<string, Ptr<SubMesh> >::iterator subIter = submeshes.begin();
		while (subIter != submeshes.end()) {
			Ptr<SubMesh> submesh = (*subIter).second;
			ss << submesh.ptr();
			subIter++;
		}

		return ss.str();
	}

	void Material::bind() {
		if (texture) {
			texture->bind();
		} else {
			if (ambient) {
				float color[4] = { ambient->r, ambient->g, ambient->b, ambient->a };
				glMaterialfv(GL_FRONT, GL_AMBIENT, color);
			}
			if (diffuse) {
				float color[4] = { diffuse->r, diffuse->g, diffuse->b, diffuse->a };
				glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
			}
			if (specular) {
				float color[4] = { specular->r, specular->g, specular->b, specular->a };
				glMaterialfv(GL_FRONT, GL_SPECULAR, color);
			}
		}
	}

	void Material::unbind() {
		if (texture) {
			texture->unbind();
		}
	}

	string Material::toString() {
		stringstream ss;

		ss << "--- MATERIAL ---" << endl;
		ss << "Name: [" << name() << "]" << endl;
		ss << "ID: [" << id() << "]" << endl;
		ss << "mapFilename: [" << mapFilename << "]" << endl;
		ss << "ambient: " << ambient.ptr() << endl;
		ss << "diffuse: " << diffuse.ptr() << endl;
		ss << "specular: " << specular.ptr() << endl;

		return ss.str();
	}
}
