#ifndef GEOMETRY_H
#define GEOMETRY_H

#include <string>
#include <vector>

#include "Texture.h"
#include "ModelReader3DS.h"
#include "Drawable.h"
#include "Printable.h"
#include "Ptr.h"
#include "PtrInterface.h"

namespace GameEngine {

	using namespace Fwk;
	using namespace std;

	class RGBColor :  public PtrInterface<RGBColor>, public Printable {
	public:
		RGBColor() : r(1), g(1), b(1), a(1) { }
		float r, g, b, a;

		string toString();
	};

	class Material : public UniqueEntity, public Printable {
	public:
		Material() : UniqueEntity(), mapFilename("") { }
		Material(string name) : UniqueEntity(name), mapFilename("") { }
		string mapFilename;
		Ptr<RGBColor> ambient;
		Ptr<RGBColor> diffuse;
		Ptr<RGBColor> specular;
		Ptr<Texture> texture;

		void bind();
		void unbind();

		string toString();
	};

	class Point : public PtrInterface<Point> {
	public:
		Point() : x(0), y(0) { }
		Point(float xx, float yy) : x(xx), y(yy) { }

		float x, y;
	};

	class Vec3 : public PtrInterface<Vec3>, public Printable {
	public:
		Vec3() : x(0), y(0), z(0) { }
		Vec3(float xx, float yy, float zz) : x(xx), y(yy), z(zz) { }
		Vec3(Ptr<Vec3> v) : x(v->x), y(v->y), z(v->z) { }
		float x, y, z;

		float mag();
		void normalize();
		Ptr<Vec3> cross(const Ptr<Vec3> v);
		Ptr<Vec3> mult(const Ptr<Vec3> v);
		Ptr<Vec3> mult(float v);
		Ptr<Vec3> plus(const Ptr<Vec3> v);
		Ptr<Vec3> sub(const Ptr<Vec3> v);

		string toString();
	};

	class Vertex3 : public PtrInterface<Vertex3>, public Printable {
	public:
		Vertex3() : x(0), y(0), z(0) { }
		Vertex3(float xx, float yy, float zz) : x(xx), y(yy), z(zz) { }
		float x, y, z;

		Ptr<Vec3> sub(const Ptr<Vertex3> v);

		string toString();
	};

	class Mapping : public PtrInterface<Mapping>, public Printable {
	public:
		Mapping() : u(0), v(0) { }
		float u, v;

		string toString();
	};

	class Face : public PtrInterface<Face>, public Printable {
	public:
		Face() : a(0), b(0), c(0), flags(0), materialName("") { }
		unsigned short a, b, c;
		unsigned short flags;
		string materialName;

		string toString();
	};

	class SubMesh : public Drawable, public Printable {
	public:
		SubMesh() : Drawable(), vertexes(), mappings(), faces(), normals() {
			material = new Material;
		}
		SubMesh(string name) : Drawable(name), vertexes(), mappings(), faces(), normals() {
			material = new Material;
		}
		vector<Ptr<Vertex3> > vertexes;
		vector<Ptr<Mapping> > mappings;
		vector<Ptr<Face> > faces;
		vector<Ptr<Vec3> > normals;
		Ptr<Material> material;

		void draw();
		void build();
		Ptr<Drawable> copy();

		string toString();

	protected:
		void drawMesh();
	};

	class Mesh : public Drawable, public Printable {
	public:
/*
		Mesh() : Drawable(), vertexes(), mappings(), faces(), normals() {
			texture = new Texture();
		}
		Mesh(string name) : Drawable(name), vertexes(), mappings(), faces(), normals() {
			texture = new Texture();
		}
		vector<Ptr<Vertex3> > vertexes;
		vector<Ptr<Mapping> > mappings;
		vector<Ptr<Face> > faces;
		vector<Ptr<Vec3> > normals;
		Ptr<Texture> texture;
*/
		Mesh() : Drawable(), submeshes() { }
		Mesh(string name) : Drawable(name), submeshes() { }

		map<string, Ptr<SubMesh> > submeshes;

		void draw();
		void build();
		Ptr<Drawable> copy();

		string toString();

	protected:
		void drawMesh();
	};

	class Matrix : public PtrInterface<Matrix> , public Printable {
	public:
		Matrix() {
			setToIdentity();
		};
		Matrix(float* m) {
			memcpy(values_, m, sizeof(values_));
		}

		void apply();

		void putValues(float *output) {
			memcpy(output, values_, sizeof(values_));
		}

		float& value(unsigned int i) { return values_[i]; }
		float value(unsigned int i) const { return values_[i]; }

		void row(unsigned int i, float* result);

		void translate(float x, float y, float z);
		void rotate(float angle, float x, float y, float z);
		void scale(float x, float y, float z);

		string toString();

		Ptr<Matrix> matrixMult(Ptr<Matrix> other);
		Ptr<Vec3> mult(Ptr<Vec3> v);

	protected:
		// indexes go top to bottom in col, then left to right in rows
		float values_[16];

		void setToIdentity();
	};

	class Math {
	public:
		static const float PI;

		static float degrees(float radians);
		static float radians(float degrees);
		static float det(float a, float b, float c, float d);
		static Ptr<Vec3> computeNormal(Ptr<Vertex3> v1, Ptr<Vertex3> v2, Ptr<Vertex3> v3);

	};

}

#endif