#include <SDL/SDL_opengl.h>
#include <algorithm>

#include "HUD.h"

namespace GameEngine {

	bool Container::contains(float x, float y) {
		return x >= x_ && x < x_ + width_ && y <= y_ && y > y_ - height_;
	}

	void HUD::setupOrthoView(int x, int y) {
		glDisable(GL_LIGHTING);
		glDisable(GL_DEPTH_TEST);
		glDepthMask(GL_FALSE);

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0, x, 0, y, -1, 1);
		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
	}

	void HUD::setupPerspectiveView() {
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		glEnable(GL_LIGHTING);
		glEnable(GL_DEPTH_TEST);
		glDepthMask(GL_TRUE);
	}

	void Container::draw() {
		if (visible_) {
			glPushMatrix();
			glTranslatef(x_ + width_ / 2.0, y_ - height_ / 2.0, 0);
			glScalef(width_, height_, 1);

			glColor3f(backgroundColor_->r, backgroundColor_->g, backgroundColor_->b);

			glBegin(GL_QUADS);
			glVertex2f(-0.5, -0.5);
			glVertex2f(0.5, -0.5);
			glVertex2f(0.5, 0.5);
			glVertex2f(-0.5, 0.5);
			glEnd();

			glPopMatrix();

			// draw children
			vector<Ptr<Container> >::iterator it = children_.begin();
			while (it != children_.end()) {
				Ptr<Container> container = *it;
				container->draw();
				it++;
			}
		}
	}

	void Container::clear() {
		vector<Ptr<Container> >::iterator it = children_.begin();
		while (it != children_.end()) {
			Ptr<Container> container = *it;
			container->onDel();
			it++;
		}
		children_.clear();
	}

	void Component::drawText() {
		if (visible_) {
			// white text
			// TODO: dynamic color
			glColor4f(1, 1, 1, 0.5);

			// where to write the text
			glRasterPos2f(textx_, texty_);

			// write each character
			std::string::iterator charIter = text_.begin();
			while (charIter != text_.end()) {
				glutBitmapCharacter(glutFont_, *charIter);
				++charIter;
			}
		}
	}

	void Component::draw() {
		if (visible_) {
			Container::draw();

			if (texture_) {
				texture_->bind();
			}

			drawText();

			if (texture_) {
				texture_->unbind();
			}
		}
	}

	int Component::textHeight() {
		if (glutFont() == GLUT_BITMAP_HELVETICA_18) {
			return 18;
		} else if (glutFont() == GLUT_BITMAP_HELVETICA_12) {
			return 12;
		} else if (glutFont() == GLUT_BITMAP_HELVETICA_10) {
			return 10;
		}
		return 0;
	}

	void HUD::mouseListenerNew(Ptr<MouseListener> listener) {
		mouseListeners_.push_back(listener);
	}

	void HUD::mouseListenerDel(Ptr<MouseListener> listener) {
		deletedMouseListeners_.push_back(listener);
	}

	void HUD::deleteDeletedMouseListeners() {
		vector<Ptr<MouseListener> >::iterator dIt = deletedMouseListeners_.begin();
		while (dIt != mouseListeners_.end()) {
			Ptr<MouseListener> listener = *dIt;

			vector<Ptr<MouseListener> >::iterator mIt;
			mIt = find(mouseListeners_.begin(), mouseListeners_.end(), listener);
			if (mIt != mouseListeners_.end()) {
				mouseListeners_.erase(mIt);
			}

			dIt++;
		}
		deletedMouseListeners_.clear();
	}

	bool HUD::onMouseDown(int sdlButton, int x, int y) {
		bool retVal = false;
		vector<Ptr<MouseListener> >::iterator it = mouseListeners_.begin();
		while (it != mouseListeners_.end()) {
			Ptr<MouseListener> listener = *it;
			retVal = retVal || listener->onMouseDown(sdlButton, x, y);
			it++;
		}
		return retVal;
	}

	bool HUD::onMouseUp(int sdlButton, int x, int y) {
		bool retVal = false;
		vector<Ptr<MouseListener> >::iterator it = mouseListeners_.begin();
		while (it != mouseListeners_.end()) {
			Ptr<MouseListener> listener = *it;
			retVal = retVal || listener->onMouseUp(sdlButton, x, y);
			it++;
		}
		return retVal;
	}

	bool HUD::onMouseMotion(int x, int y) {
		bool retVal = false;
		vector<Ptr<MouseListener> >::iterator it = mouseListeners_.begin();
		while (it != mouseListeners_.end()) {
			Ptr<MouseListener> listener = *it;
			retVal = retVal || listener->onMouseMotion(x, y);
			it++;
		}
		return retVal;
	}

}
