#include <SDL/SDL_opengl.h>
#include <string>
#include <iostream>
#include <sstream>
#include <algorithm>

#include "Ptr.h"
#include "Model.h"

namespace GameEngine {

	using namespace std;

	void Model::build() {

		// can't have nested glNewList() calls
		for (int i = 0; i < meshes_.size(); i++) {
			meshes_[i]->build();
		}

		if (drawId_ > 0) {
			glDeleteLists(drawId_, 1);
		}
		drawId_ = glGenLists(1);
		glNewList(drawId_, GL_COMPILE);

		for (int i = 0; i < meshes_.size(); i++) {
			meshes_[i]->draw();
		}

		glEndList();

	}

	void Model::draw() {
		// for selecting objects in 3d
		glPushName(id_);

		// draw
		//glEnable(GL_LIGHTING);
		glColor4f(1, 1, 1, 1);
		for (int i = 0; i < meshes_.size(); i++) {
			meshes_[i]->draw();
		}
		//glCallList(drawId_);
		//glDisable(GL_LIGHTING);

		// for selecting objects in 3d
		glPopName();
	}

	Ptr<Drawable> Model::copy() {
		return new Model;
	}

	string Model::toString() {
		stringstream ss;

		ss << "*** MODEL ***" << endl;
		ss << "Name: [" << name() << "]" << endl;
		ss << "ID: [" << id() << "]" << endl;

		vector<Ptr<Mesh> >::iterator meshIter = meshes_.begin();
		map<string, Ptr<Material> >::iterator matIter = materials_.begin();

		ss << "*************************************" << endl;
		ss << "Meshes:" << endl;
		ss << "*************************************" << endl;
		while (meshIter != meshes_.end()) {
			ss << (*meshIter).ptr() << endl;
			meshIter++;
		}
		ss << "*************************************" << endl;
		ss << "Materials:" << endl;
		ss << "*************************************" << endl;
		while (matIter != materials_.end()) {
			ss << (*matIter).second.ptr() << endl;
			matIter++;
		}
		ss << "*************************************" << endl;

		return ss.str();
	}

	void Model::normalize() {
		// compute the bounding box
		Ptr<Vec3> minn = new Vec3;
		Ptr<Vec3> maxx = new Vec3;

		vector<Ptr<Mesh> >::iterator meshIter = meshes_.begin();
		while (meshIter != meshes_.end()) {
			Ptr<Mesh> mesh = *meshIter;
			map<string, Ptr<SubMesh> >::iterator subIter = mesh->submeshes.begin();
			while (subIter != mesh->submeshes.end()) {
				Ptr<SubMesh> submesh = (*subIter).second;
				vector<Ptr<Vertex3> >::iterator vertIter = submesh->vertexes.begin();
				while (vertIter != submesh->vertexes.end()) {
					Ptr<Vertex3> v = *vertIter;

					if (meshIter == meshes_.begin() &&
						subIter == mesh->submeshes.begin() &&
						vertIter == submesh->vertexes.begin()) {
						// initialize min and max
						minn->x = maxx->x = v->x;
						minn->y = maxx->y = v->y;
						minn->z = maxx->z = v->z;
					} else {
						minn->x = min(v->x, minn->x);
						minn->y = min(v->y, minn->y);
						minn->z = min(v->z, minn->z);

						maxx->x = max(v->x, maxx->x);
						maxx->y = max(v->y, maxx->y);
						maxx->z = max(v->z, maxx->z);
					}

					vertIter++;
				}
				subIter++;
			}
			meshIter++;
		}

		// scale so largest side = 1
		float scale = 1.0;
		float xside = maxx->x - minn->x;
		float yside = maxx->y - minn->y;
		float zside = maxx->z - minn->z;
		float longestSide = max(max(xside, yside), zside);
		if (longestSide > 0) {
			scale = 1.0 / longestSide;
		}
		normTransform_->scale(scale, scale, scale);

		// move to center
		Ptr<Vec3> centerTrans = new Vec3;
		centerTrans->x = -1 * (minn->x + ((maxx->x - minn->x) / 2.0f));
		centerTrans->y = -1 * (minn->y + ((maxx->y - minn->y) / 2.0f));
		centerTrans->z = -1 * (minn->z + ((maxx->z - minn->z) / 2.0f));
		normTransform_->translate(centerTrans->x, centerTrans->y, centerTrans->z);
	}

	vector<Ptr<Material> > Model::loadMaterials() {
		vector<Ptr<Material> > materials;

		map<string, Ptr<Material> >::iterator matIter = materials_.begin();
		while (matIter != materials_.end()) {
			Ptr<Material> mat = (*matIter).second;

			Ptr<Texture> tex = new Texture(mat->name());
			if (tex->loadTexture(mat->mapFilename)) {
				mat->texture = tex;
				materials.push_back(mat);
			} else {
				mat->texture = new Texture;
			}

			matIter++;
		}

		// set the texture reference for each submesh
		vector<Ptr<Mesh> >::iterator meshIter = meshes_.begin();
		while (meshIter != meshes_.end()) {
			Ptr<Mesh> mesh = *meshIter;
			map<string, Ptr<SubMesh> >::iterator subIter = mesh->submeshes.begin();
			while (subIter != mesh->submeshes.end()) {
				Ptr<SubMesh> submesh = (*subIter).second;
				if (submesh->faces.size() > 0) {
					submesh->material = materials_[submesh->faces[0]->materialName];
				}
				subIter++;
			}
			meshIter++;
		}

		return materials;
	}
}