#include "ModelReader3DS.h"
#include "ModelLoader.h"
#include "Ptr.h"

namespace GameEngine {

	Ptr<Model> ModelLoader::load3DSFile(string filename) {
		Ptr<M3DS::ModelReader3DS> reader = new M3DS::ModelReader3DS;
		reader->process3ds(filename);

		// convert M3DS data to GameEngine format
		Ptr<Model> model = new Model;

		vector<Ptr<M3DS::ObjectBlock> > blocks = reader->objectBlocks();
		map<string, Ptr<M3DS::Material> > materials = reader->materials();

		vector<Ptr<M3DS::ObjectBlock> >::iterator blockIter = blocks.begin();
		map<string, Ptr<M3DS::Material> >::iterator matIter = materials.begin();

		while (blockIter != blocks.end()) {
			Ptr<M3DS::ObjectBlock> block = *blockIter;
			model->meshNew(meshFrom3ds(block->mesh, block->name));
			blockIter++;
		}

		while (matIter != materials.end()) {
			Ptr<Material> mat = materialFrom3ds((*matIter).second);
			model->materialNew((*matIter).first, mat);
			matIter++;
		}

		// compile the model into a display list
		model->build();

		// generate the normalization matrix that puts the model in the center
		// and scales it to unit size
		model->normalize();

		return model;
	}

	Ptr<Vertex3> ModelLoader::vertexFrom3ds(M3DS::Vertex v3ds) {
		Ptr<Vertex3> v = new Vertex3;
		v->x = v3ds.x;
		v->y = v3ds.y;
		v->z = v3ds.z;
		return v;
	}

	Ptr<Mapping> ModelLoader::mappingFrom3ds(M3DS::Mapping m3ds) {
		Ptr<Mapping> m = new Mapping;
		m->u = m3ds.u;
		m->v = m3ds.v;
		return m;
	}

	Ptr<Face> ModelLoader::faceFrom3ds(M3DS::Face f3ds) {
		Ptr<Face> f = new Face;
		// TODO: verify vertices are specified in wrong direction
		//f->a = f3ds.a;
		//f->b = f3ds.b;
		//f->c = f3ds.c;
		f->a = f3ds.c;
		f->b = f3ds.b;
		f->c = f3ds.a;
		f->flags = f3ds.flags;
		f->materialName = f3ds.materialName;
		return f;
	}
	
	Ptr<Mesh> ModelLoader::meshFrom3ds(M3DS::Mesh m3ds, string name) {
		Ptr<Mesh> m = new Mesh(name);

		vector<M3DS::Vertex> verts3ds = m3ds.vertexes;
		vector<M3DS::Mapping> maps3ds = m3ds.mappings;
		vector<M3DS::Face> faces3ds = m3ds.faces;

		vector<M3DS::Vertex>::iterator vIter = verts3ds.begin();
		vector<M3DS::Mapping>::iterator mIter = maps3ds.begin();
		vector<M3DS::Face>::iterator fIter = faces3ds.begin();

		vector<Ptr<Vertex3> > vertexes(verts3ds.size());
		vector<Ptr<Mapping> > mappings(maps3ds.size());

		unsigned int i = 0;
		while (vIter != verts3ds.end()) {
			vertexes[i] = vertexFrom3ds(*vIter);
			vIter++;
			i++;
		}
		i = 0;
		while (mIter != maps3ds.end()) {
			mappings[i] = mappingFrom3ds(*mIter);
			mIter++;
			i++;
		}

		while (fIter != faces3ds.end()) {
			Ptr<Face> face = faceFrom3ds(*fIter);
			if (m->submeshes.find(face->materialName) == m->submeshes.end()) {
				Ptr<SubMesh> submesh = new SubMesh;
				m->submeshes[face->materialName] = submesh;
			}
			Ptr<SubMesh> submesh = m->submeshes[face->materialName];

			submesh->vertexes = vertexes;
			submesh->mappings = mappings;
			submesh->faces.push_back(faceFrom3ds(*fIter));
			submesh->normals.push_back(normal(submesh, submesh->faces.back()));

			fIter++;
		}
/*
//		m->vertexes.resize(verts3ds.size());
//		m->mappings.resize(maps3ds.size());
//		m->faces.resize(faces3ds.size());
//		m->normals.resize(faces3ds.size());

		unsigned int i = 0;
		while (vIter != verts3ds.end()) {
			m->vertexes[i] = vertexFrom3ds(*vIter);
			vIter++;
			i++;
		}
		i = 0;
		while (mIter != maps3ds.end()) {
			m->mappings[i] = mappingFrom3ds(*mIter);
			mIter++;
			i++;
		}
		i = 0;
		while (fIter != faces3ds.end()) {
			m->faces[i] = faceFrom3ds(*fIter);
			m->normals[i] = normal(m, m->faces[i]);
			fIter++;
			i++;
		}
*/
		return m;
	}

	Ptr<Vec3> ModelLoader::normal(Ptr<SubMesh> submesh, Ptr<Face> face) {
		return Math::computeNormal(
			submesh->vertexes[face->a], submesh->vertexes[face->b], submesh->vertexes[face->c]);
	}

	Ptr<RGBColor> ModelLoader::rgbFrom3ds(M3DS::RGB c3ds) {
		Ptr<RGBColor> c = new RGBColor;
		c->r = c3ds.r;
		c->g = c3ds.g;
		c->b = c3ds.b;
		return c;
	}

	Ptr<Material> ModelLoader::materialFrom3ds(Ptr<M3DS::Material> m3ds) {
		Ptr<Material> m = new Material;
		m->nameIs(m3ds->name);
		m->mapFilename = m3ds->mapFilename;
		m->ambient = rgbFrom3ds(m3ds->ambient);
		m->diffuse = rgbFrom3ds(m3ds->diffuse);
		m->specular = rgbFrom3ds(m3ds->specular);
		return m;
	}

}