#ifndef PARTICLE_H
#define PARTICLE_H

#include <vector>
#include <ctime> 

#include "Ptr.h"
#include "PtrInterface.h"
#include "Color.h"
#include "Geometry.h"
#include "SceneNode.h"
#include "CustomSceneNodes.h"
#include "Engine.h"

namespace GameEngine {

	using namespace Fwk;
	using namespace std;

	//
	// RandomNumberGenerator
	//

	class RandomNumberGenerator : public PtrInterface<RandomNumberGenerator> {
	public:
		static Ptr<RandomNumberGenerator> Instance() {
			if (!instance_) { 
				instance_ = new RandomNumberGenerator;
			}
			return instance_;
		}
		float gen() { return 1.0f * rand() / RAND_MAX; }
		float neg1to1() { return 1 - gen() * 2; }
	protected:
		RandomNumberGenerator() {
			srand(time(0));
		}

		static Ptr<RandomNumberGenerator> instance_;
	};

	//
	// Particle
	//

	class Particle : public Drawable {
	public:
		Particle() :
		  Drawable(),
		  startLife_(0),
		  life_(0),
		  size_(0),
		  color_(Color4::Black()),
		  velocity_(new Vec3),
		  texture_(NULL),
		  sceneNode_(NULL),
		  billboardTrue_(true) {
		}

		void del();

		float lifePercent() { return startLife_ > 0 ? life_ / startLife_ : 0; }
		bool dead() { return life_ < 0; }
		float size() { return size_; }
		Ptr<Color4> color() { return color_; }
		Ptr<Vertex3> position() { return sceneNode_ ? sceneNode_->worldPosition() : NULL; }
		Ptr<Vec3> velocity() { return velocity_; }
		Ptr<Texture> texture() { return texture_; }
		Ptr<SceneNode> sceneNode() { return sceneNode_; }

		void lifeIs(float life) { startLife_ = life_ = life; }
		void lifeDec(float v) { life_ -= v; }
		void positionIs(Ptr<Vertex3> pos) { sceneNode_->moveTo(pos->x, pos->y, pos->z); }
		void sizeIs(float size) { size_ = size; sceneNode_->scale(size, size, size); }
		void colorIs(Ptr<Color4> color) { color_ = color; }
		void velocityIs(Ptr<Vec3> velocity) { velocity_ = velocity; }
		void textureIs(Ptr<Texture> texture) { texture_ = texture; }
		void sceneNodeIs(Ptr<SceneNode> sceneNode) { sceneNode_ = sceneNode; }

		virtual void build() { }
		virtual void draw();

		void billboardToggle() { billboardTrue_ = !billboardTrue_; }
		void billboardIs(bool v) { billboardTrue_ = v; }

	protected:
		float startLife_;
		float life_;
		float size_;
		Ptr<Color4> color_;
		Ptr<Vec3> velocity_;
		Ptr<Texture> texture_;
		Ptr<SceneNode> sceneNode_;
		bool billboardTrue_;
	};

	//
	// ParticleSystem
	//

	class ParticleSystem : public Drawable {
	public:
		//static const float Z_INC;

		ParticleSystem(Ptr<Engine> engine, Ptr<Vertex3> origin, Ptr<SceneNode> parentSN) :
		  Drawable(),
		  particles_(),
		  sceneNode_(new SceneNode),
		  engine_(engine),
		  origin_(origin),
		  billboardTrue_(true) {
			parentSN->childNew(sceneNode_);
		}

		virtual void build() { }
		virtual void draw();

		virtual void init() = 0;
		virtual void update(float elapsedMillis) = 0;

		bool dead() { return particles_.size() == 0; }
		void del();

		void billboardToggle() {
			billboardTrue_ = !billboardTrue_;
			for (int i = 0; i < particles_.size(); i++) {
				particles_[i]->billboardToggle();
			}
		}
		void billboardIs(bool v) {
			billboardTrue_ = v;
			for (int i = 0; i < particles_.size(); i++) {
				particles_[i]->billboardIs(v);
			}
		}
		int particleNum() { return particles_.size(); }

	protected:
		vector<Ptr<Particle> > particles_;
		Ptr<SceneNode> sceneNode_;
		Ptr<Engine> engine_;
		Ptr<Vertex3> origin_;
		//static float zlevel_;
		//static int totalNumParticles_;
		bool billboardTrue_;

		void removeDeadParticles();
	};

}

#endif