#ifndef READER3DS_H
#define READER3DS_H

#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <vector>
#include <map>

#include "Ptr.h"
#include "PtrInterface.h"

namespace M3DS {

	typedef unsigned char byte;
	typedef unsigned short int word;
	typedef unsigned long int dword;

	typedef signed char int8;
	typedef signed short int int16;
	typedef signed long int int32;

	#define malloc_(type,amount) ((type*)malloc(sizeof(type)*(amount)));

	#define MAX_VERTICES	(65536)
	#define MAX_OBJBLOCKS	(256)

	#define	DUMP3DS_INDENT	(4)

	typedef struct chunk_name {
		word id;
		char* name;
	} chunk_name;

	#pragma pack(2)
	struct chunk_header {
		word    id;
		dword   len;
	};
	#pragma pack()

	/* Enumeration of chunk IDs to be used by chunk lists */
	enum {
		CHUNK_NULL		= 0x0000,
		CHUNK_RGBF      = 0x0010,
		CHUNK_RGBB      = 0x0011,
		CHUNK_M3D_VERSION = 0x0002,
		//CHUNK_RBGB2     = 0x0012,       // ?? NOT HLS.

		CHUNK_MAIN      = 0x4D4D,
			CHUNK_OBJMESH   = 0x3D3D,
				CHUNK_MESH_VERSION	= 0x3d3e,
				CHUNK_MASTER_SCALE	= 0x100,
				CHUNK_BKGCOLOR  = 0x1200,
				CHUNK_AMBCOLOR  = 0x2100,
				CHUNK_OBJBLOCK  = 0x4000,
					CHUNK_TRIMESH   = 0x4100,
						CHUNK_VERTLIST  = 0x4110,
						CHUNK_FACELIST  = 0x4120,
						CHUNK_FACEMAT   = 0x4130,
						CHUNK_MAPLIST   = 0x4140,
						CHUNK_SMOOLIST  = 0x4150,
						CHUNK_TRMATRIX  = 0x4160,
					CHUNK_LIGHT     = 0x4600,
						CHUNK_SPOTLIGHT = 0x4610,
					CHUNK_CAMERA    = 0x4700,
			CHUNK_MATERIAL  = 0xAFFF,
				CHUNK_MATNAME   = 0xA000,
				CHUNK_AMBIENT   = 0xA010,
				CHUNK_DIFFUSE   = 0xA020,
				CHUNK_SPECULAR  = 0xA030,
				CHUNK_TEXTURE   = 0xA200,
				CHUNK_BUMPMAP   = 0xA230,
				CHUNK_MAPFILE   = 0xA300,
			CHUNK_KEYFRAMER = 0xB000,
				CHUNK_KFHDR		= 0xb00a,
				CHUNK_FRAMES    = 0xB008,
					CHUNK_KFCURTIME	= 0xb009,
					CHUNK_OBJECT_NODE_TAG	= 0xb002,
					CHUNK_CAMERA_NODE_TAG	= 0xb003,
					CHUNK_TARGET_NODE_TAG	= 0xb004,
						CHUNK_BOUNDBOX		= 0xb014,
						CHUNK_NODE_ID		= 0xB030,
  						CHUNK_NODE_HDR 		= 0xB010,
						CHUNK_PIVOT			= 0xB013,
						CHUNK_INSTANCE_NAME	= 0xb011,
						CHUNK_POS_TRACK_TAG	= 0xB020,
						CHUNK_ROT_TRACK_TAG	= 0xB021,
						CHUNK_SCL_TRACK_TAG	= 0xB022,
						CHUNK_FOV_TRACK_TAG = 0xB023,
						CHUNK_ROL_TRACK_TAG = 0xB024

	};

	#define NAME_BUFFER_SIZE 256

	#define M3D_REPORT_ERROR(f) { \
		printf("Error in pos %ld\n", ftell(f)); \
	}

	#define M3D_REPORT_CHUNK_DATA(indent, verbose, args) { \
		if(verbose) { \
			printf("%*c",indent, ' '); \
			printf args; \
			(void)putchar('\n'); \
		} \
	}

	#define M3D_FREAD(i,f) fread(&i,sizeof(i), 1, f)

	#define M3D_FREAD_EXT(i,f) { \
		if (M3D_FREAD(i,f) != 1) { \
			M3D_REPORT_ERROR(f); \
			return; \
		} \
	}

	using namespace Fwk;
	using namespace std;

	class Reader3DS : public PtrInterface<Reader3DS> {
	public:
		void process3ds(string filename);

	protected:
		Reader3DS() :
			verbose(false) {
			initChunkNames();
		}

		void initChunkNames();
		void m3dFreadName(FILE* f);
		string chunkName(int id);
		void chunkReader(FILE* f, unsigned int indent, unsigned long p);
		void runChunkFunc(word id, FILE* f, unsigned int indent, unsigned long p);

		virtual void chunkM3dVersionReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkRGBFReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkRGBBReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkMainReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkObjMeshReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkMeshVersionReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkMasterScaleReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkBkgColorReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkAmbColorReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkObjBlockReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkTriMeshReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkVertListReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkFaceListReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkFaceMatReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkMapListReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkSmooListReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkTrMatrixReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkLightReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkSpotlightReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkCameraReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkMaterialReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkMatNameReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkAmbientReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkDiffuseReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkSpecularReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkTextureReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkBumpMapReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkMapFileReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkKeyFramerReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkKfHdrReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkFramesReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkKfCurTimeReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkObjectNodeTagReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkCameraNodeTagReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkTargetNodeTagReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkNodeIdReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkNodeHdrReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkPivotReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkInstanceNameReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkPosTrackTagReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkRotTrackTagReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkSclTrackTagReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkFovTrackTagReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkRolTrackTagReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkBoundBoxReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkNullReader(FILE* f, unsigned int indent, unsigned long p) { };
		virtual void chunkUnknownReader(FILE* f, unsigned int indent, unsigned long p) { };

		bool verbose;
		map<int, string> chunkNames;
		char nameBuffer[NAME_BUFFER_SIZE];
	};

	class PrintReader3DS : public Reader3DS {
	public:
		PrintReader3DS() : Reader3DS() { verbose = true; }

		void chunkM3dVersionReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkRGBFReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkRGBBReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkMeshVersionReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkMasterScaleReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkObjBlockReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkVertListReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkFaceListReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkFaceMatReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkMapListReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkSmooListReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkTrMatrixReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkLightReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkSpotlightReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkCameraReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkMatNameReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkMapFileReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkKfHdrReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkFramesReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkKfCurTimeReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkNodeIdReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkNodeHdrReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkPivotReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkInstanceNameReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkPosTrackTagReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkRotTrackTagReader(FILE* f, unsigned int indent, unsigned long p);
		void chunkSclTrackTagReader(FILE* f, unsigned int indent, unsigned long p);
	};

}

#endif
