#include <vector>
#include <math.h>
#include <GLee/GLee.h>
#include <SDL/SDL_opengl.h>
#include <GL/glut.h>
#include <iostream>

#include "Ptr.h"
#include "Terrain.h"
#include "Geometry.h"
#include "TextureManager.h"
#include "ShaderManager.h"

namespace GameEngine {

	using namespace std;

	Ptr<Vec3> TerrainCell::enemyMoveVec() {
		switch (enemyMoveDir_) {
			case North: return new Vec3(0, 1, 0);
			case South: return new Vec3(0, -1, 0);
			case East: return new Vec3(1, 0, 0);
			case West: return new Vec3(-1, 0, 0);
			default: return new Vec3(0, 0, 0);
		}
	}

	void Terrain::initMaterials() {
		Ptr<Texture> t;
		if (grassMat_ == NULL) {
			grassMat_ = new Material;
			t = TextureManager::Instance()->texture(TextureManager::GRASS_IMAGE_FILE);
			grassMat_->texture = t;
		}
		if (dirtMat_ == NULL) {
			dirtMat_ = new Material;
			t = TextureManager::Instance()->texture(TextureManager::DIRT_IMAGE_FILE);
			dirtMat_->texture = t;
		}
	}

	void Terrain::resize(unsigned int height, unsigned int width) {
		height_ = height;
		width_ = width;

		initMaterials();

		cells_.clear();
		for (int row = 0; row < height_; row++) {
			vector<Ptr<TerrainCell> > rowVec(width_);
			cells_.push_back(rowVec);
			for (int col = 0; col < width_; col++) {
				Ptr<TerrainCell> tc = new TerrainCell;
				if (tc->cellType() == TerrainCell::GrassCellType) {
					tc->materialIs(grassMat_);
					tc->bumpMapIs(TextureManager::Instance()->texture(
						TextureManager::GRASS_BUMPMAP_IMAGE_FILE));
				} else if (tc->cellType() == TerrainCell::DirtCellType) {
					tc->materialIs(dirtMat_);
					tc->bumpMapIs(TextureManager::Instance()->texture(
						TextureManager::DIRT_BUMPMAP_IMAGE_FILE));
				}
				cells_[row][col] = tc;
			}
		}
	}

	Ptr<Point> Terrain::cellRowCol(unsigned int cellId) {
		// first, check to see if cellId is in table
		if (cellId >= cells_[0][0]->id() &&
			cellId <= cells_[height_ - 1][width_ - 1]->id()) {

			// find the id distance between each cell
			unsigned int idDiff = cells_[0][1]->id() - cells_[0][0]->id();

			// now, find the index
			unsigned int flatIndex = (cellId - cells_[0][0]->id()) / idDiff;
			unsigned int row = flatIndex / width_;
			unsigned int col = flatIndex % height_;

			return new Point(row, col);
		}
		return NULL;
	}

	Ptr<Point> Terrain::cellOffset(unsigned int cellId) {
		Ptr<Point> rowCol = cellRowCol(cellId);
		if (rowCol != NULL) {
			return cellOffset(rowCol->x, rowCol->y);
		}
		return NULL;
	}

	Ptr<Point> Terrain::cellOffset(int row, int col) {
		float x = 0.5 + -1.0 * CELL_WIDTH * width() / 2.0 + CELL_WIDTH * col * 1.0;
		float y = 0.5 + 1.0 * CELL_HEIGHT * height() / 2.0 - CELL_HEIGHT * row * 1.0;
		return new Point(x, y);
	}

	Ptr<TerrainCell> Terrain::cell(float x, float y) {
		//float x = 0.5 + -1.0 * CELL_WIDTH * width() / 2.0 + CELL_WIDTH * col * 1.0;
		//float y = 0.5 + 1.0 * CELL_HEIGHT * height() / 2.0 - CELL_HEIGHT * row * 1.0;
		float colf = (x + CELL_WIDTH * width() / 2.0) / (CELL_WIDTH * 1.0);
		float rowf = -1.0 * (y - CELL_HEIGHT * height() / 2.0) / (CELL_HEIGHT * 1.0);
		rowf++;
		if (colf >= 0 && colf <= width() && rowf >= 0 && rowf <= height() - 1) {
			int col = colf;
			int row = rowf;
			return cells_[row][col];
		}
		return NULL;
	}

	Ptr<TerrainCell> Terrain::cell(unsigned int cellId) {
		Ptr<Point> rowCol = cellRowCol(cellId);
		if (rowCol != NULL) {
			return cells_[rowCol->x][rowCol->y];
		}
		return NULL;
	}

	void Terrain::select(unsigned int cellId, bool placeable) {
		selectedCell_ = cell(cellId);
		selectedCellPlaceable_ = placeable;
	}

	void Terrain::cellDataIs(int row, int col, TerrainCell::CellType cellType, TerrainCell::Direction dir) {
		Ptr<TerrainCell> tc = cells_[row][col];
		tc->enemyMoveDirIs(dir);
		tc->cellTypeIs(cellType);
		if (tc->cellType() == TerrainCell::GrassCellType) {
			tc->materialIs(grassMat_);
			tc->bumpMapIs(TextureManager::Instance()->texture(TextureManager::GRASS_BUMPMAP_IMAGE_FILE));
		} else if (tc->cellType() == TerrainCell::DirtCellType) {
			tc->materialIs(dirtMat_);
			tc->bumpMapIs(TextureManager::Instance()->texture(TextureManager::DIRT_BUMPMAP_IMAGE_FILE));
		}
	}

	void Terrain::build() { }

	void Terrain::draw() {
		float cellCenteredX = CELL_WIDTH / 2.0;
		float cellCenteredY = CELL_HEIGHT / 2.0;
		float cellx, celly;
		Ptr<Point> offset;

		Ptr<Shader> shader = ShaderManager::Instance()->shader(ShaderManager::BUMPMAP_SHADER);
		for (int row = 0; row < height(); row++) {
			for (int col = 0; col < width(); col++) {
				offset = cellOffset(row, col);
				cellx = offset->x;
				celly = offset->y;

				Ptr<TerrainCell> cell = cells_[row][col];

				glPushMatrix();
				glTranslatef(cellx, celly, 0);

				glColor4f(1, 1, 1, 1);

				if (cell->material()) {
					if (renderLevel_ == RLFull) {
						shader->bind();

						unsigned int tx0 = glGetUniformLocationARB(shader->program(), "colorMap");
						glUniform1iARB(tx0, 0);
						glActiveTextureARB(GL_TEXTURE0_ARB);
						cell->material()->bind();

						unsigned int tx1 = glGetUniformLocationARB(shader->program(), "normalMap");
						glUniform1iARB(tx1, 1);
						glActiveTextureARB(GL_TEXTURE1_ARB);
						cell->bumpMap()->bind();

						unsigned int invLoc = glGetUniformLocationARB(shader->program(), "invRadius");
						glUniform1fARB(invLoc, 0.0001);

						int mode = shader->mode() % 3;
						unsigned int modeLoc = glGetUniformLocationARB(shader->program(), "mode");
						glUniform1fARB(modeLoc, mode);
					} else {
						//cell->material()->bind();
					}
				}

				glPushName(cell->id());

				glBegin(GL_QUADS);
				glNormal3f(0, 0, 1);
				glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0, 0);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0, 0);
				glVertex3f(-cellCenteredX, -cellCenteredY, 0);

				glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1, 0);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1, 0);
				glVertex3f(cellCenteredX, -cellCenteredY, 0);

				glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 1, 1);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 1, 1);
				glVertex3f(cellCenteredX, cellCenteredY, 0);

				glMultiTexCoord2fARB(GL_TEXTURE0_ARB, 0, 1);
				glMultiTexCoord2fARB(GL_TEXTURE1_ARB, 0, 1);
				glVertex3f(-cellCenteredX, cellCenteredY, 0);
				glEnd();

				glPopName();

				if (cell->material()) {
					if (renderLevel_ == RLFull) {
						glActiveTextureARB(GL_TEXTURE0_ARB);
						cell->material()->unbind();
						glActiveTextureARB(GL_TEXTURE1_ARB);
						cell->bumpMap()->unbind();
					} else {
						//cell->material()->unbind();
					}
				}

				if (renderLevel_ == RLFull) {
					shader->unbind();
				}

				if (selectedCell_ == cell && renderLevel_ == RLFull) {
					glPushMatrix();
					glDisable(GL_LIGHTING);

					glEnable(GL_BLEND);
					glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

					if (selectedCellPlaceable_) {
						glColor4f(1.0f, 1.0f, 1.0f, 0.3f);
					} else {
						glColor4f(1.0f, 0.0f, 0.0f, 0.3f);
					}
					glTranslatef(0, 0, 0.01f);

					glBegin(GL_QUADS);
					glNormal3f(0, 0, 1);
					glVertex3f(-cellCenteredX, -cellCenteredY, 0);
					glVertex3f(cellCenteredX, -cellCenteredY, 0);
					glVertex3f(cellCenteredX, cellCenteredY, 0);
					glVertex3f(-cellCenteredX, cellCenteredY, 0);
					glEnd();

					glDisable(GL_BLEND);
					glEnable(GL_LIGHTING);
					glPopMatrix();
				}

				glPopMatrix();
			}
		}
	}

	unsigned int Terrain::placeObject(Ptr<SceneNode> object, unsigned int cellId) {
		Ptr<Point> offset;

		Ptr<TerrainCell> selCell = cell(cellId);
		for (int row = 0; row < cells_.size(); row++) {
			for (int col = 0; col < cells_[row].size(); col++) {
				Ptr<TerrainCell> cell = cells_[row][col];
				if (cell->id() == cellId && !cell->occupieeId()) {
					offset = cellOffset(row, col);
					float transx = offset->x;
					float transy = offset->y;

					// assuming object is at (0, 0, 0)
					object->translate(transx, transy, 0);
					
					// TODO: get actual model id, not scene node id
					cell->occupieeIdIs(object->id());

					return cell->occupieeId();
				}
			}
		}
		return 0;
	}

	void Terrain::removeObject(unsigned int cellId) {
		for (int row = 0; row < cells_.size(); row++) {
			for (int col = 0; col < cells_[row].size(); col++) {
				Ptr<TerrainCell> cell = cells_[row][col];
				if (cell->id() == cellId) {
					cell->occupieeIdIs(0);
				}
			}
		}
	}

}
