#ifndef TERRAIN_H
#define TERRAIN_H

#include <string>

#include "SceneNode.h"
#include "Geometry.h"
#include "Drawable.h"
#include "UniqueEntity.h"
#include "Ptr.h"
#include "PtrInterface.h"

namespace GameEngine {

	using namespace Fwk;
	using namespace std;

	class TerrainCell : public UniqueEntity {
	public:
		enum Direction { None, North, South, East, West };
		enum CellType { GrassCellType, DirtCellType };

		TerrainCell() : 
		  UniqueEntity(),
		  cellType_(GrassCellType),
		  occupieeId_(0),
		  enemyMoveDir_(None) {
		  material_ = NULL;
		  bumpMap_ = NULL;
		}

		TerrainCell(unsigned int type) : 
			UniqueEntity(),
			cellType_(type),
			occupieeId_(0),
			enemyMoveDir_(None) {
			material_ = NULL;
			bumpMap_ = NULL;
		}

		unsigned int cellType() { return cellType_; }
		unsigned int occupieeId() { return occupieeId_; }
		Ptr<Material> material() { return material_; }
		Ptr<Texture> bumpMap() { return bumpMap_; }
		Direction enemyMoveDir() { return enemyMoveDir_; }
		Ptr<Vec3> enemyMoveVec();

		void cellTypeIs(unsigned int cellType) { cellType_ = cellType; }
		void occupieeIdIs(unsigned int occupieeId) { occupieeId_ = occupieeId; }
		void materialIs(Ptr<Material> material) { material_ = material; }
		void bumpMapIs(Ptr<Texture> bumpMap) { bumpMap_ = bumpMap; }
		void enemyMoveDirIs(Direction dir) { enemyMoveDir_ = dir; }

	protected:
		unsigned int cellType_;
		unsigned int occupieeId_;
		Ptr<Material> material_;
		Ptr<Texture> bumpMap_;

		// TODO: abstract out of generic TerrainCell class
		Direction enemyMoveDir_;
	};

	class Terrain : public Drawable {
	public:
		static const int CELL_WIDTH = 1;
		static const int CELL_HEIGHT = 1;

		enum RenderLevel { RLSimple, RLFull };

		Terrain() :
		  Drawable("Terrain"),
		  height_(0),
		  width_(0),
		  cells_(),
		  selectedCellPlaceable_(true),
		  sceneNode_(NULL),
		  renderLevel_(RLFull) {
			selectedCell_ = NULL;
			grassMat_ = NULL;
			dirtMat_ = NULL;
		}
		Terrain(unsigned int height, unsigned int width) :
		  Drawable("Terrain"),
		  height_(height),
		  width_(width),
		  cells_(),
		  selectedCellPlaceable_(true),
		  sceneNode_(NULL),
		  renderLevel_(RLFull) {
			selectedCell_ = NULL;
			for (int row = 0; row < height_; row++) {
				vector<Ptr<TerrainCell> > rowVec(width_);
				cells_.push_back(rowVec);
				for (int col = 0; col < width_; col++) {
					cells_[row][col] = new TerrainCell;
				}
			}
			grassMat_ = NULL;
			dirtMat_ = NULL;
		}

		void draw();
		void build();

		unsigned int placeObject(Ptr<SceneNode> object, unsigned int cellId);
		void removeObject(unsigned int cellId);

		unsigned int width() { return width_; }
		unsigned int height() { return height_; }
		Ptr<SceneNode> sceneNode() { return sceneNode_; }

		void sceneNodeIs(Ptr<SceneNode> sn) { sceneNode_ = sn; }
		void renderLevelIs(RenderLevel rl) { renderLevel_ = rl; }

		void resize(unsigned int height, unsigned int width);

		Ptr<Point> cellRowCol(unsigned int cellId);
		Ptr<Point> cellOffset(unsigned int cellId);
		Ptr<Point> cellOffset(int row, int col);
		Ptr<TerrainCell> cell(unsigned int cellId);
		Ptr<TerrainCell> cell(float x, float y);
		void select(unsigned int cellId, bool placeable);

		void cellDataIs(int row, int col, TerrainCell::CellType cellType, TerrainCell::Direction dir);

	protected:
		unsigned int width_;
		unsigned int height_;
		vector<vector<Ptr<TerrainCell> > > cells_;
		Ptr<TerrainCell> selectedCell_;
		bool selectedCellPlaceable_;
		Ptr<SceneNode> sceneNode_;
		RenderLevel renderLevel_;

		// TODO: move to resource manager
		Ptr<Material> grassMat_;
		Ptr<Material> dirtMat_;
		void initMaterials();

		void drawTerrain();		  
	};
}

#endif
