#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>
#include <SDL/SDL_image.h>
#include <iostream>
#include <sstream>

#include "Texture.h"

namespace GameEngine {

	using namespace std;

	SDL_Surface *Texture::loadImage(string filename) {

		/* Thanks for Lazy Foo's tutorial on image loading in SDL:
		 * http://lazyfoo.net/SDL_tutorials/lesson03/windows/msvsnet05e/index.php
		 */

		//The image that's loaded
		SDL_Surface* loadedImage = NULL;

		//The optimized image that will be used
		SDL_Surface* optimizedImage = NULL;

		//Load the image using
		loadedImage = IMG_Load(filename.c_str());

		//If the image loaded
		if(loadedImage != NULL) {
			//Create an optimized image
			optimizedImage = SDL_DisplayFormat(loadedImage);

			//Free the old image
			SDL_FreeSurface(loadedImage);
		}

		//Return the optimized image
		return optimizedImage;
	}

	unsigned int Texture::storeGlTexture(SDL_Surface *surface) {

		/*	Implementation 1:
		 *	http://gpwiki.org/index.php/SDL:Tutorials:Using_SDL_with_OpenGL
		 */

		unsigned int texture;
		GLenum textureFormat;
		GLint nOfColors;

		if (surface) { 
			// Check that the image's width is a power of 2
			if ((surface->w & (surface->w - 1)) != 0) {
				cout << "warning: image's width is not a power of 2" << endl;
			}

			// Also check if the height is a power of 2
			if ((surface->h & (surface->h - 1)) != 0 ) {
				cout << "warning: image's height is not a power of 2" << endl;
			}

			// get the number of channels in the SDL surface
			nOfColors = surface->format->BytesPerPixel;
			if (nOfColors == 4) {     // contains an alpha channel
				if (surface->format->Rmask == 0x000000ff)
					textureFormat = GL_RGBA;
				else
					textureFormat = GL_BGRA;
/*
				// set alpha channel to full
				int flatIndex;
				for (int row = 0; row < surface->h; row++) {
					for (int col = 0; col < surface->w; col++) {
						flatIndex = row * surface->h * 4 + col * 4 + 3;
						((char *) surface->pixels)[flatIndex] = 255;
					}
				}
*/
			} else if (nOfColors == 3) {     // no alpha channel
				if (surface->format->Rmask == 0x000000ff)
					textureFormat = GL_RGB;
				else
					textureFormat = GL_BGR;
			} else {
				cout << "warning: the image is not truecolor...this will probably break" << endl;
				// this error should not go unhandled
			}

			// Have OpenGL generate a texture object handle for us
			glGenTextures(1, &texture);

			// Bind the texture object
			glBindTexture(GL_TEXTURE_2D, texture);

			// Set the texture's stretching properties
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);

			gluBuild2DMipmaps(
				GL_TEXTURE_2D,
				3,
				surface->w,
				surface->h,
				textureFormat,
				GL_UNSIGNED_BYTE,
				surface->pixels);
/*
			// Edit the texture object's image data using the information SDL_Surface gives us
			glTexImage2D(
				GL_TEXTURE_2D,
				0,
				nOfColors,
				surface->w,
				surface->h,
				0,
				textureFormat,
				GL_UNSIGNED_BYTE,
				surface->pixels);
*/
		} else {
			cout << "SDL could not load image: " << SDL_GetError() << endl;
		}    

		// Free the SDL_Surface only if it was successfully created
		if (surface) { 
			SDL_FreeSurface(surface);
		}

		return texture;

		/*	Implementation 2:
		 *	http://steinsoft.net/index.php?site=Programming/Code%20Snippets/SDL/no1
		 */
/*
		// Standard OpenGL texture creation code
		glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
		 		
		glGenTextures(1, &texture);
		glBindTexture(GL_TEXTURE_2D, texture);
		 
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magFilter);
		 		
		gluBuild2DMipmaps(GL_TEXTURE_2D,3,bmpFile->w,
		   bmpFile->h,GL_BGR_EXT,
		   GL_UNSIGNED_BYTE,bmpFile->pixels);
		 
		//Free surface after using it
		SDL_FreeSurface(bmpFile);
*/
	}

	unsigned int Texture::loadTexture(string filename) {
		unsigned int texId = 0;
		if (filename.substr(0, 8) != "textures") {
			filename = "textures/" + filename;
		}
		SDL_Surface *image = loadImage(filename);
		if (image) {
			texId = storeGlTexture(image);
			if (texId > 0) {
				filename_ = filename;
				texId_ = texId;
			}
		}
		return texId;
	}

	void Texture::bind() {
		if (texId_ > 0) {
			//glEnable(GL_LIGHTING);
			glEnable(GL_COLOR_MATERIAL);
			glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
			glColor4f(1.0, 1.0, 1.0, 1.0);

			glEnable(GL_TEXTURE_2D);
			glBindTexture(GL_TEXTURE_2D, texId_);
		}
	}

	void Texture::unbind() {
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_COLOR_MATERIAL);
	}

	string Texture::toString() {
		stringstream ss;
		ss << "Texture: name=[" << filename_ << "] id=[" << texId_ << "]";
		return ss.str();
	}
}
