#ifndef TOWER_H
#define TOWER_H

#include <string>

#include "Ptr.h"
#include "PtrInterface.h"
#include "Geometry.h"
#include "SceneNode.h"
#include "Copiable.h"
#include "Color.h"
#include "Terrain.h"

namespace Tower {

	using namespace Fwk;
	using namespace std;
	using namespace GameEngine;

	class Enemy;
	class Projectile;

	//
	// Data classes
	//

	class TowerUpgradeData : public PtrInterface<TowerUpgradeData>, public Copiable<TowerUpgradeData> {
	public:
		TowerUpgradeData() :
		  cost_(0),
		  range_(0),
		  damage_(0) { }

		int cost() { return cost_; }
		int range() { return range_; }
		int damage() { return damage_; }

		void costIs(int cost) { cost_ = cost; }
		void rangeIs(int range) { range_ = range; }
		void damageIs(int damage) { damage_ = damage; }

		virtual Ptr<TowerUpgradeData> copy();

	protected:
		int cost_;
		int range_;
		int damage_;
	};

	class TowerData : public PtrInterface<TowerData>, public Copiable<TowerData> {
	public:
		// TODO: read in tower types and tower attr's from data file
		enum TowerType { NoTowerType, FireTowerType, IceTowerType, EarthTowerType };

		TowerData(TowerType type) :
		  type_(type),
		  cost_(DEFAULT_COST),
		  range_(DEFAULT_RANGE),
		  damage_(DEFAULT_DAMAGE),
		  fireDelay_(0),
		  maxFireDelay_(DEFAULT_FIRING_DELAY),
		  hudImage_(NULL),
		  upgradeLevel_(0),
		  upgrades_() {
			initUpgrades(type);
		}

		TowerType towerType() { return type_; }
		int cost() { return cost_; }
		int range() { return range_; }
		int damage() { return damage_; }
		float maxFireDelay() { return maxFireDelay_; }
		float fireDelay() { return fireDelay_; }
		Ptr<Texture> hudImage() { return hudImage_; }
		int upgradeLevel() { return upgradeLevel_; }
		Ptr<TowerUpgradeData> upgradeNextData() {
			if (upgradeLevel_ < upgrades_.size()) {
				return upgrades_[upgradeLevel_];
			}
			return NULL;
		}

		void towerTypeIs(TowerType type) { type_ = type; }
		void costIs(int cost) { cost_ = cost; }
		void rangeIs(int range) { range_ = range; }
		void damageIs(int damage) { damage_ = damage; }
		void fireDelayDec(float elapsedMillis) { fireDelay_ -= elapsedMillis; }
		void fireDelayReset() { fireDelay_ = maxFireDelay_; }
		void maxFireDelayIs(float maxFireDelay) { maxFireDelay_ = maxFireDelay; }
		void hudImageIs(Ptr<Texture> hudImage) { hudImage_ = hudImage; }
		void upgradeLevelIs(int upgradeLevel) { upgradeLevel_ = upgradeLevel; }
		void upgradeNew(Ptr<TowerUpgradeData> upgradeData) { upgrades_.push_back(upgradeData); }

		void upgrade();

		virtual Ptr<TowerData> copy();

	protected:
		static const float DEFAULT_COST;
		static const float DEFAULT_RANGE;
		static const float DEFAULT_DAMAGE;
		static const float DEFAULT_FIRING_DELAY;
		static const int FIRE_L1_COST;
		static const int FIRE_L2_COST;
		static const int FIRE_L3_COST;
		static const int FIRE_L1_DAMAGE;
		static const int FIRE_L2_DAMAGE;
		static const int FIRE_L3_DAMAGE;
		static const int FIRE_L1_RANGE;
		static const int FIRE_L2_RANGE;
		static const int FIRE_L3_RANGE;
		static const int ICE_L1_COST;
		static const int ICE_L2_COST;
		static const int ICE_L3_COST;
		static const int ICE_L1_DAMAGE;
		static const int ICE_L2_DAMAGE;
		static const int ICE_L3_DAMAGE;
		static const int ICE_L1_RANGE;
		static const int ICE_L2_RANGE;
		static const int ICE_L3_RANGE;
		static const int EARTH_L1_COST;
		static const int EARTH_L2_COST;
		static const int EARTH_L3_COST;
		static const int EARTH_L1_DAMAGE;
		static const int EARTH_L2_DAMAGE;
		static const int EARTH_L3_DAMAGE;
		static const int EARTH_L1_RANGE;
		static const int EARTH_L2_RANGE;
		static const int EARTH_L3_RANGE;

		TowerType type_;
		int cost_;
		int range_;
		int damage_;
		float fireDelay_;
		float maxFireDelay_;
		Ptr<Texture> hudImage_;
		int upgradeLevel_;
		vector<Ptr<TowerUpgradeData> > upgrades_;

		void initUpgrades(TowerType towerType);
	};

	class EnemyData : public PtrInterface<EnemyData>, public Copiable<EnemyData> {
	public:
		static const float DEFAULT_HEALTH;
		static const float DEFAULT_GOLD;

		enum EnemyType { CalEnemyType, UCLAEnemyType, USCEnemyType };

		EnemyData(EnemyType enemyType) :
		  enemyType_(enemyType),
		  gold_(DEFAULT_GOLD),
		  health_(DEFAULT_HEALTH),
		  speed_(DEFAULT_SPEED) { }

		EnemyType enemyType() { return enemyType_; }
		float gold() { return gold_; }
		float health() { return health_; }
		float speed() { return speed_; }

		void goldIs(float gold) { gold_ = gold; }
		void healthDec(float v) { health_ -= v; }
		void healthIs(int health) { health_ = health; }
		void speedIs(float speed) { speed_ = speed; }

		virtual Ptr<EnemyData> copy();

	protected:
		static const float DEFAULT_SPEED;

		EnemyType enemyType_;
		int gold_;
		int health_;
		float speed_;
	};

	//
	// Non-data classes
	//

	class Tower : public Drawable, public Copiable<Tower> {
	public:
		Tower(TowerData::TowerType towerType) :
		  Drawable(),
		  towerData_(new TowerData(towerType)),
		  sceneNode_(NULL),
		  terrainCell_(NULL) { }

		void del();

		virtual void build() { };
		virtual void draw();

		virtual Ptr<Tower> copy();

		void towerDataIs(Ptr<TowerData> towerData) { towerData_ = towerData; }
		void sceneNodeIs(Ptr<SceneNode> sceneNode) { sceneNode_ = sceneNode; }
		void terrainCellIs(Ptr<TerrainCell> tc) { terrainCell_ = tc; }
		Ptr<TerrainCell> terrainCell() { return terrainCell_; }
		Ptr<Vertex3> position();

		Ptr<TowerData> towerData() { return towerData_; }
		TowerData::TowerType towerType() { return towerData_->towerType(); }
		unsigned int range() { return towerData_->range(); }
		unsigned int damage() { return towerData_->damage(); }
		int firingDelay() { return towerData_->fireDelay(); }
		void firingDelayDec(int elapsedMillis) { towerData_->fireDelayDec(elapsedMillis); }
		void resetFiringDelay() { towerData_->fireDelayReset(); }

	protected:
		Ptr<TowerData> towerData_;
		Ptr<SceneNode> sceneNode_;
		Ptr<TerrainCell> terrainCell_;
	};

	class Projectile : public Drawable, public Copiable<Projectile> {
	public:
		Projectile() :
			Drawable(),
			tower_(NULL),
			damage_(0),
			dead_(false) {
			target_ = NULL;
			velocity_ = NULL;
			sceneNode_ = NULL;
			tower_ = NULL;
		}

		Projectile(Ptr<Tower> tower) :
			Drawable(),
			tower_(tower),
			damage_(0),
			dead_(false) {
			target_ = NULL;
			velocity_ = NULL;
			sceneNode_ = NULL;
		}

		virtual void build() { };
		virtual void draw();

		virtual Ptr<Projectile> copy();

		void del();

		unsigned int damage() { return damage_; }
		Ptr<Enemy> target() { return target_; }
		Ptr<Tower> tower() { return tower_; }
		TowerData::TowerType projectileType() { return tower_->towerType(); }
		bool dead() { return dead_; }
		Ptr<Vec3> velocity() { return velocity_; }

		void towerIs(Ptr<Tower> tower) { tower_ = tower; }
		void damageIs(unsigned int damage) { damage_ = damage; }
		void deadIs(bool v) { dead_ = v; }
		void targetIs(Ptr<Enemy> target) { target_ = target; }
		void velocityIs(Ptr<Vec3> velocity) { velocity_ = velocity; }
		void sceneNodeIs(Ptr<SceneNode> sceneNode) { sceneNode_ = sceneNode; }

		void updatePosition(float elapsedMillis);
		Ptr<Vertex3> position();
		bool intersectsTarget();
		void seekTarget();

		// TODO: temp only
		float radius() { return DEFAULT_RADIUS; }

	protected:
		static const float DEFAULT_RADIUS;
		static const float DEFAULT_SPEED;

		Ptr<Tower> tower_;
		unsigned int damage_;
		bool dead_;
		Ptr<Enemy> target_;
		Ptr<Vec3> velocity_;
		Ptr<SceneNode> sceneNode_;
	};

	class Enemy : public Drawable, public Copiable<Enemy> {
	public:
		static const float DEFAULT_VELOCITY;
		static const float DEFAULT_RADIUS;

		Enemy(EnemyData::EnemyType enemyType) :
		  Drawable(),
		  enemyData_(new EnemyData(enemyType)),
		  sceneNode_(NULL) {
			velocity_ = new Vec3(0, - enemyData_->speed(), 0);
			initMaterial();
		}

		virtual void build() { };
		virtual void draw();

		virtual Ptr<Enemy> copy();

		void del();
		void initMaterial();

		Ptr<EnemyData> enemyData() { return enemyData_; }
		int health() { return enemyData_->health(); }
		float speed() { return enemyData_->speed(); }
		Ptr<Vec3> velocity() { return velocity_; }
		Ptr<Material> material() { return material_; }

		void enemyDataIs(Ptr<EnemyData> enemyData) { enemyData_ = enemyData; }
		void velocityIs(Ptr<Vec3> velocity) { velocity_ = velocity; }
		void sceneNodeIs(Ptr<SceneNode> sceneNode) { sceneNode_ = sceneNode; }
		void materialIs(Ptr<Material> material) { material_ = material; }

		void updatePosition(float elapsedMillis);
		Ptr<Vertex3> position();
		void moveTo(Ptr<Vertex3> point) { sceneNode_->moveTo(point->x, point->y, point->z); }

		bool getSmacked(Ptr<Projectile> projectile);

		// TODO: temp only
		float radius() { return DEFAULT_RADIUS; }

	protected:
		Ptr<EnemyData> enemyData_;
		Ptr<Vec3> velocity_;
		Ptr<SceneNode> sceneNode_;
		Ptr<Material> material_;
	};

	class Light : public Drawable {
	public:
		Light(unsigned int lightId) :
		  Drawable(),
		  lightId_(lightId),
		  sphereColor_(Color4::Yellow()),
		  speed_(0.05),
		  angle_(0),
		  hide_(false) { }

	  void sceneNodeIs(Ptr<SceneNode> sceneNode) { sceneNode_ = sceneNode; }
	  Ptr<SceneNode> sceneNode() { return sceneNode_; }

	  virtual void build() { }
	  virtual void draw();
	  void hideToggle() { 
		  hide_ = !hide_;
	  }

	  void update(float millis) { 
		  angle_ += speed_ * millis;
	  }

	protected:
		unsigned int lightId_;
		Ptr<Color4> sphereColor_;
		Ptr<SceneNode> sceneNode_;
		float speed_;
		float angle_;
		bool hide_;
	};

}

#endif
