#include <string>
#include <iostream>
#include <map>

#include "Model.h"
#include "ModelLoader.h"
#include "CustomSceneNodes.h"
#include "TowerEngine.h"
#include "TowerHUD.h"
#include "Tower.h"
#include "CustomSceneNodes.h"
#include "Timer.h"

namespace Tower {

	const float TowerEngineData::DEFAULT_GOLD = 50;
	const float TowerEngineData::DEFAULT_LIVES = 10;
	const float TowerEngineData::DEFAULT_LEVEL = 1;
	const int TowerEngine::ENEMY_START_ROW = -1;
	const int TowerEngine::ENEMY_START_COL = 2;
	const unsigned int TowerEngine::NUM_ENEMIES = 20;

	// called once every draw loop
	void TowerEngine::update(float elapsedMillis) {
		map<unsigned int, Ptr<Tower> >::iterator towerIter;
		map<unsigned int, Ptr<Enemy> >::iterator enemyIter;
		map<unsigned int, Ptr<Projectile> >::iterator projIter;

		// update light
		light_->update(elapsedMillis);

		if (isGameOver()) {
			gameState_ = GSWaitingForInput;
			reset();
			Ptr<TowerHUD> thud = ptr_cast<TowerHUD>(hud_);
			thud->onRoundOver();
			thud->contextInfoReset();
		} else if (isRoundOver()) {
			gameState_ = GSWaitingForInput;
			towerEngineData_->levelInc();
			Ptr<TowerHUD> thud = ptr_cast<TowerHUD>(hud_);
			thud->onRoundOver();
		}

		// update all enemy positions
		unsigned int escapedEnemyId = 0;
		enemyIter = enemies_.begin();
		while (enemyIter != enemies_.end()) {
			Ptr<Enemy> enemy = (*enemyIter).second;

			// get current cell
			Ptr<TerrainCell> tc = terrain_->cell(enemy->position()->x, enemy->position()->y);
			if (tc != NULL) {
				Ptr<Point> cellCenterPoint = terrain_->cellOffset(tc->id());
				Ptr<Vertex3> cellCenter =
					new Vertex3(cellCenterPoint->x, cellCenterPoint->y, enemy->position()->z);

				// check to see if we've crossed the middle of the cell
				Ptr<Vertex3> oldPos = enemy->position();
				enemy->updatePosition(elapsedMillis);
				Ptr<Vertex3> newPos = enemy->position();

				Ptr<Vec3> centerToOld = cellCenter->sub(oldPos);
				Ptr<Vec3> centerToNew = cellCenter->sub(newPos);
				// multiply together and see if sign of x or y is + or -
				Ptr<Vec3> c = centerToOld->mult(centerToNew);

				if (centerToNew->x == 0 && centerToNew->y == 0) {
					// new pos is exactly on center, update enemy velocity
					enemy->velocityIs(tc->enemyMoveVec()->mult(enemy->speed()));
				} else if (c->x < 0 || c->y < 0) {
					// the center was crossed
					enemy->velocityIs(tc->enemyMoveVec()->mult(enemy->speed()));

					// fix the position
					float newDist = centerToNew->mag();
					Ptr<Vec3> moveDir = enemy->velocity()->mult(newDist);
					Ptr<Vertex3> movePoint =
						new Vertex3(cellCenter->x + moveDir->x, cellCenter->y + moveDir->y, cellCenter->z);
					enemy->moveTo(movePoint);
				}
			} else if (enemy->position()->x > terrain_->width() / 2.0) {
				escapedEnemyId = enemy->id();
			} else {
				enemy->updatePosition(elapsedMillis);
			}

			enemyIter++;
		}

		if (escapedEnemyId > 0) {
			// enemy got out!
			towerEngineData_->livesDec();
			// remove enemy scene node
			enemies_[escapedEnemyId]->del();
			// remove from enemies list
			enemies_.erase(escapedEnemyId);
		}

		// update all projectile positions
		projIter = projectiles_.begin();
		while (projIter != projectiles_.end()) {
			Ptr<Projectile> proj = (*projIter).second;
			proj->updatePosition(elapsedMillis);
			projIter++;
		}

		// update all particle systems
		int i = 0;
		while (i < particleSystems_.size()) {
			Ptr<ParticleSystem> psys = particleSystems_[i];
			if (psys->dead()) {
				psys->del();
				particleSystems_.erase(particleSystems_.begin() + i);
			} else {
				psys->update(elapsedMillis);
				i++;
			}
		}

		// check for collisions
		vector<unsigned int> deadProjectileIds;
		vector<unsigned int> deadEnemyIds;

		projIter = projectiles_.begin();
		while (projIter != projectiles_.end()) {
			Ptr<Projectile> proj = (*projIter).second;

			if (proj->intersectsTarget()) {
				// mark projectile for removal
				proj->deadIs(true);
				deadProjectileIds.push_back((*projIter).first);

				// decrease enemy stats
				Ptr<Enemy> enemy = proj->target();
				bool enemyDead = enemy->getSmacked(proj);
				if (enemyDead) {
					deadEnemyIds.push_back(proj->target()->id());
				}

				// init particle system
				Ptr<ParticleSystem> psys = NULL;
				if (proj->projectileType() == TowerData::EarthTowerType) {
					psys = new DustSystem(this, proj, rootSN_);
					psys->init();
					psys->billboardIs(billboardTrue_);
					particleSystems_.push_back(psys);
				} else if (proj->projectileType() == TowerData::FireTowerType) {
					psys = new FireSystem(this, proj, rootSN_);
					psys->init();
					psys->billboardIs(billboardTrue_);
					particleSystems_.push_back(psys);
				} else if (proj->projectileType() == TowerData::IceTowerType) {
					psys = new IceSystem(this, proj->position(), rootSN_);
				}
			}
			projIter++;
		}

		// remove collided items
		vector<unsigned int>::iterator pIter = deadProjectileIds.begin();
		while (pIter != deadProjectileIds.end()) {
			int index = *pIter;
			Ptr<Projectile> proj = projectiles_[index];

			// remove projectile scene node
			proj->del();

			// remove from tower engine projectile list
			projectiles_.erase(index);

			pIter++;
		}

		// take action on dead enemies
		vector<unsigned int>::iterator eIter = deadEnemyIds.begin();
		while (eIter != deadEnemyIds.end()) {
			int index = *eIter;
			if (enemies_.find(index) != enemies_.end()) {
				Ptr<Enemy> enemy = enemies_[index];

				// remove enemy scene node
				enemy->del();

				// remove from enemies list
				enemies_.erase(index);

				// update stats
				towerEngineData_->goldInc(enemy->enemyData()->gold());
				ptr_cast<TowerHUD>(hud_)->updateTowerAndUpgradeButtons(true);
			}

			eIter++;
		}

		// towers fire if in range
		// (right now, n^2 method...need n log n)
		towerIter = towers_.begin();
		while (towerIter != towers_.end()) {
			Ptr<Tower> tower = (*towerIter).second;

			tower->firingDelayDec(elapsedMillis);

			// first, check if we can even fire
			if (tower->firingDelay() <= 0) {

				// enemies are sorted in order, so fire at first one in range
				enemyIter = enemies_.begin();
				while (enemyIter != enemies_.end()) {
					Ptr<Enemy> enemy = (*enemyIter).second;

					// check range
					Ptr<Vertex3> tPos = tower->position();
					Ptr<Vertex3> ePos = enemy->position();
					Ptr<Vec3> distVec = tPos->sub(ePos);
					float dist = distVec->mag();
					if (dist < tower->range()) {

						// FIRE!!
						Ptr<Projectile> p = projectileNewInstance(tower->towerType());
						p->towerIs(tower);
						Ptr<DrawableSceneNode> sn = new DrawableSceneNode(p);
						sn->translate(tPos->x, tPos->y, 0.5);
						rootSN_->childNew(sn);
						p->sceneNodeIs(sn);
						p->damageIs(tower->damage());
						p->targetIs(enemy);
						p->seekTarget();
						projectiles_[p->id()] = p;

						// add particle trails
						if (p->projectileType() == TowerData::IceTowerType) {
							Ptr<ParticleSystem> psys = new IceTrailSystem(this, p, rootSN_);
							psys->init();
							psys->billboardIs(billboardTrue_);
							particleSystems_.push_back(psys);
						}

						tower->resetFiringDelay();
						break;
					}
					enemyIter++;
				}
			}
			towerIter++;
		}
	}

	void TowerEngine::addTower(Ptr<Tower> tower, unsigned int cellId) {
		Ptr<DrawableSceneNode> sn = new DrawableSceneNode(tower);
		terrain_->placeObject(sn, cellId);
		rootSN_->childNew(sn);
		tower->sceneNodeIs(sn);
		towers_[tower->id()] = tower;

		// update stats
		towerEngineData_->goldDec(tower->towerData()->cost());

		if (towerEngineData_->gold() < tower->towerData()->cost()) {
			ptr_cast<TowerHUD>(hud_)->updateTowerAndUpgradeButtons(false);
		}
	}

	void TowerEngine::startRound() {
		if (enemyType_ == EnemyData::CalEnemyType) {
			enemyType_ = EnemyData::UCLAEnemyType;
		} else if (enemyType_ == EnemyData::UCLAEnemyType) {
			enemyType_ = EnemyData::USCEnemyType;
		} else if (enemyType_ == EnemyData::USCEnemyType) {
			enemyType_ = EnemyData::CalEnemyType;
		}
		initEnemies(enemyType_);
		gameState_ = GSRunning;
	}

	bool TowerEngine::isGameOver() {
		return towerEngineData_->lives() == 0 && gameState_ == GSRunning;
	}
	bool TowerEngine::isRoundOver() {
		return enemies_.size() == 0 && gameState_ == GSRunning;
	}

	void TowerEngine::reset() {
		removeTowers();
		removeEnemies();
		removeProjectiles();
		removeParticles();
		towerEngineData_->reset();
		Ptr<TowerHUD> thud = ptr_cast<TowerHUD>(hud_);
		thud->onReset();
		thud->updateTowerAndUpgradeButtons(true);
	}

	void TowerEngine::onPick(unsigned int pickedId) {
		// check cell or tower selected
		Ptr<TowerHUD> towerHud = ptr_cast<TowerHUD>(hud_);

		Ptr<TerrainCell> cell = terrain_->cell(pickedId);
		Ptr<Tower> t = tower(pickedId);

		if (cell && cell->cellType() == TerrainCell::GrassCellType) {
			if (towerHud->newTowerType() != TowerData::NoTowerType) {
				t = towerNewInstance(towerHud->newTowerType());
				t->terrainCellIs(cell);
				if (t->towerData()->cost() <= towerEngineData_->gold()) {
					// add tower to scene
					addTower(t, pickedId);
					// reset tower info in HUD
					towerHud->newTowerTypeIs(TowerData::NoTowerType);
					towerHud->contextInfoReset();
					if (towerEngineData_->gold() < t->towerData()->cost()) {
						towerHud->updateTowerAndUpgradeButtons(false);
					}
				}
				terrain_->select(0, false);
			}
		} else if (t) {
			terrain_->select(0, false);
			towerEngineData_->selectedTowerIs(t);
			towerHud->contextTowerUpgradeInfo(t);
			towerHud->newTowerTypeIs(TowerData::NoTowerType);
		}
	}

	void TowerEngine::onMouseMotion(unsigned int x, unsigned int y) {
		Engine::onMouseMotion(x, y);

		bool cellpicked = false;
		Ptr<TowerHUD> thud = ptr_cast<TowerHUD>(hud_);
		if (thud->newTowerType() != TowerData::NoTowerType) {
			unsigned int pickedId = processPick(true, x, y);
			Ptr<TerrainCell> tc = terrain_->cell(pickedId);
			if (tc) {
				cellpicked = true;

				// check to see if tower on cell
				bool towerincell = false;
				map<unsigned int, Ptr<Tower> >::iterator ti = towers_.begin();
				while (ti != towers_.end()) {
					Ptr<Tower> tower = (*ti).second;
					if (tc == tower->terrainCell()) {
						terrain_->select(0, false);
						towerincell = true;
						break;
					}
					ti++;
				}

				if (!towerincell) {
					bool placeable = tc->cellType() == TerrainCell::GrassCellType;
					terrain_->select(pickedId, placeable);
				}
			}
		}
		if (!cellpicked) {
			terrain_->select(0, false);
		}
	}

	void TowerEngine::onKeyDown(SDLKey key, SDLMod modifer) {
		Engine::onKeyDown(key, modifer);

		switch (key) {
			case SDLK_r:
				// used to be startRound code
				break;
			case SDLK_l:
				light_->hideToggle();
				break;
			case SDLK_f: {
				billboardTrue_ = !billboardTrue_;
				for (int i = 0; i < particleSystems_.size(); i++) {
					particleSystems_[i]->billboardToggle();
				}
				break;
			}
			default:
				break;
		}
	}

	void TowerEngine::removeEnemies() {
		for (int i = 0; i < enemies_.size(); i++) {
			Ptr<Enemy> e = enemies_[i];
			if (e) {
				e->del();
			}
		}
		enemies_.clear();
	}

	void TowerEngine::removeTowers() {
		for (int i = 0; i < towers_.size(); i++) {
			Ptr<Tower> t = towers_[i];
			if (t) {
				t->del();
			}
		}
		towers_.clear();
	}

	void TowerEngine::removeProjectiles() {
		for (int i = 0; i < projectiles_.size(); i++) {
			Ptr<Projectile> p = projectiles_[i];
			if (p) {
				p->del();
			}
		}
		projectiles_.clear();
	}

	void TowerEngine::removeParticles() {
		for (int i = 0; i < particleSystems_.size(); i++) {
			Ptr<ParticleSystem> p = particleSystems_[i];
			if (p) {
				p->del();
			}
		}
		particleSystems_.clear();
	}

	void TowerEngine::initTemplates() {
		towerTemplates_[TowerData::FireTowerType] = new Tower(TowerData::FireTowerType);
		towerTemplates_[TowerData::IceTowerType] = new Tower(TowerData::IceTowerType);
		towerTemplates_[TowerData::EarthTowerType] = new Tower(TowerData::EarthTowerType);

		enemyTemplates_[EnemyData::CalEnemyType] = new Enemy(EnemyData::CalEnemyType);
		enemyTemplates_[EnemyData::UCLAEnemyType] = new Enemy(EnemyData::UCLAEnemyType);
		enemyTemplates_[EnemyData::USCEnemyType] = new Enemy(EnemyData::USCEnemyType);

		projectileTemplates_[TowerData::FireTowerType] = new Projectile();
		projectileTemplates_[TowerData::IceTowerType] = new Projectile();
		projectileTemplates_[TowerData::EarthTowerType] = new Projectile();
	}

	Ptr<Tower> TowerEngine::towerNewInstance(TowerData::TowerType towerType) {
		return towerTemplates_[towerType]->copy();
	}

	Ptr<Enemy> TowerEngine::enemyNewInstance(EnemyData::EnemyType enemyType) {
		return enemyTemplates_[enemyType]->copy();
	}

	Ptr<Projectile> TowerEngine::projectileNewInstance(TowerData::TowerType towerType) {
		return projectileTemplates_[towerType]->copy();
	}

	void TowerEngine::initEnemies(EnemyData::EnemyType enemyType) {

		static int level = 0;
		level++;

		removeEnemies();

		// init enemies
		for (int i = 0; i < NUM_ENEMIES; i++) {
			Ptr<Enemy> e = enemyNewInstance(enemyType);
			e->enemyData()->healthIs(EnemyData::DEFAULT_HEALTH * 2 * level);
			e->enemyData()->goldIs(EnemyData::DEFAULT_GOLD * 2 * level);

			Ptr<DrawableSceneNode> sn = new DrawableSceneNode(e);
			Ptr<Point> offset = terrain_->cellOffset(ENEMY_START_ROW - i, ENEMY_START_COL);
			sn->translate(offset->x, offset->y, Enemy::DEFAULT_RADIUS);

			rootSN_->childNew(sn);

			e->sceneNodeIs(sn);

			enemies_[e->id()] = e;
		}
	}

	void TowerEngine::initMyStuff() {
		hud_ = new TowerHUD(this, WINDOW_WIDTH, WINDOW_HEIGHT);
		rootSN_->translate(-3, 0, -8);

		terrain_->resize(14, 14);

		// TODO: read in from data file
		terrain_->cellDataIs(0, 2, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(1, 2, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(2, 2, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(3, 2, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(4, 2, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(4, 3, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(4, 4, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(4, 5, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(4, 6, TerrainCell::DirtCellType, TerrainCell::North);
		terrain_->cellDataIs(3, 6, TerrainCell::DirtCellType, TerrainCell::North);
		terrain_->cellDataIs(2, 6, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(2, 7, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(2, 8, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(2, 9, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(2, 10, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(3, 10, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(4, 10, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(5, 10, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(6, 10, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(7, 10, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(8, 10, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(9, 10, TerrainCell::DirtCellType, TerrainCell::West);
		terrain_->cellDataIs(9, 9, TerrainCell::DirtCellType, TerrainCell::West);
		terrain_->cellDataIs(9, 8, TerrainCell::DirtCellType, TerrainCell::West);
		terrain_->cellDataIs(9, 7, TerrainCell::DirtCellType, TerrainCell::West);
		terrain_->cellDataIs(9, 6, TerrainCell::DirtCellType, TerrainCell::North);
		terrain_->cellDataIs(8, 6, TerrainCell::DirtCellType, TerrainCell::North);
		terrain_->cellDataIs(7, 6, TerrainCell::DirtCellType, TerrainCell::North);
		terrain_->cellDataIs(6, 6, TerrainCell::DirtCellType, TerrainCell::West);
		terrain_->cellDataIs(6, 5, TerrainCell::DirtCellType, TerrainCell::West);
		terrain_->cellDataIs(6, 4, TerrainCell::DirtCellType, TerrainCell::West);
		terrain_->cellDataIs(6, 3, TerrainCell::DirtCellType, TerrainCell::West);
		terrain_->cellDataIs(6, 2, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(7, 2, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(8, 2, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(9, 2, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(10, 2, TerrainCell::DirtCellType, TerrainCell::South);
		terrain_->cellDataIs(11, 2, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(11, 3, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(11, 4, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(11, 5, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(11, 6, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(11, 7, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(11, 8, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(11, 9, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(11, 10, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(11, 11, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(11, 12, TerrainCell::DirtCellType, TerrainCell::East);
		terrain_->cellDataIs(11, 13, TerrainCell::DirtCellType, TerrainCell::East);

		//initEnemies(enemyType_);

		//modelTest();
	}
/*
	void TowerEngine::modelTest() {
		string fn1 = "3DS/PLANT.3ds";
		string fn2 = "3DS/POTTERY.3ds";
		string fn3 = "3DS/WAGON.3ds";
		string fn4 = "3DS/jeep.3ds";
		string fn5 = "3DS/CAR.3DS";

		//rootSN_->childNew(new SphereSceneNode);

		//Ptr<CubeSceneNode> csn = new CubeSceneNode;
		//rootSN_->childNew(csn);

		Ptr<ModelLoader> loader = new ModelLoader;
		//Ptr<Model> model1 = loader->load3DSFile(fn1);
		Ptr<Model> model2 = loader->load3DSFile(fn2);
		//Ptr<Model> model3 = loader->load3DSFile(fn3);
		//Ptr<Model> model4 = loader->load3DSFile(fn4);
		//Ptr<Model> model5 = loader->load3DSFile(fn5);

		//models_.push_back(model1);
		models_.push_back(model2);
		//models_.push_back(model3);
		//models_.push_back(model4);
		//models_.push_back(model5);

		//Ptr<ModelSceneNode> sn1 = new ModelSceneNode(model1);
		Ptr<ModelSceneNode> sn2 = new ModelSceneNode(model2);
		model2->sceneNodeIs(sn2);
		//Ptr<ModelSceneNode> sn3 = new ModelSceneNode(model3);
		//Ptr<ModelSceneNode> sn4 = new ModelSceneNode(model4);
		//Ptr<CubeSceneNode> sn5 = new CubeSceneNode;
		//Ptr<CubeSceneNode> sn6 = new CubeSceneNode;
		//Ptr<SphereSceneNode> sn7 = new SphereSceneNode;
		//Ptr<ModelSceneNode> sn8 = new ModelSceneNode(model5);
		//Ptr<LightSceneNode> sn9 = new LightSceneNode();

		//sn1->translate(-6, 0, 0);
		//sn1->scale(3, 3, 3);
		//sn1->rotate(90, -1, 0, 0);
		//rootSN_->childNew(sn1);
		//sn1->translate(-3, 0, 0);
		sn2->translate(3, 0, 3);
		sn2->translate(100, 0, 0);
		sn2->scale(3, 3, 3);
		rootSN_->childNew(sn2);
		//sn3->translate(0, 0, 0);
		//sn3->scale(8, 8, 8);
		//rootSN_->childNew(sn3);
		//sn4->translate(3, 0, 0);
		//rootSN_->childNew(sn4);
		//sn5->translate(6, 0, 0);
		//sn5->scale(0.5, 0.5, 0.5);
		//sn5->rotate(180, 1, 2, 3);
		//sn6->translate(2, 0, 0);
		//sn6->rotate(180, 1, 2, 3);
		//sn5->childNew(sn6);
		//rootSN_->childNew(sn5);
		//rootSN_->childNew(sn7);
		//sn8->scale(8, 8, 8);
		//sn8->translate(2, 0, 0);
		//rootSN_->childNew(sn8);
		//sn9->translate(3, 0, 0);
		//rootSN_->childNew(sn9);

		//cout << model1.ptr() << endl;
	}
*/
}