#ifndef TOWERENGINE_H
#define TOWERENGINE_H

#include <map>

#include "Ptr.h"
#include "PtrInterface.h"
#include "Engine.h"
#include "Tower.h"
#include "TowerParticle.h"

namespace Tower {

	using namespace GameEngine;
	using namespace Fwk;
	using namespace std;

	class TowerEngineData : public PtrInterface<TowerEngineData> {
	public:
		TowerEngineData() :
		  gold_(DEFAULT_GOLD),
		  lives_(DEFAULT_LIVES),
		  level_(DEFAULT_LEVEL),
		  selectedTower_(NULL) { }

		int gold() { return gold_; }
		int lives() { return lives_; }
		int level() { return level_; }
		Ptr<Tower> selectedTower() { return selectedTower_; }

		void goldInc(int v) { gold_ += v; }
		void goldDec(int v) { gold_ -= v; }
		void goldReset() { gold_ = DEFAULT_GOLD; }
		void livesDec() { lives_--; }
		void livesReset() { lives_ = DEFAULT_LIVES; }
		void levelInc() { level_++; }
		void levelReset() { level_ = DEFAULT_LEVEL; }
		void selectedTowerIs(Ptr<Tower> tower) { selectedTower_ = tower; }
		void reset() {
			goldReset();
			livesReset();
			levelReset();
		}

	protected:
		static const float DEFAULT_GOLD;
		static const float DEFAULT_LIVES;
		static const float DEFAULT_LEVEL;

		int gold_;
		int lives_;
		int level_;
		Ptr<Tower> selectedTower_;
	};

	class TowerEngine : public Engine {
	public:
		enum GameState { GSRunning, GSWaitingForInput };

		TowerEngine() :
		  Engine(),
		  towerTemplates_(),
		  enemyTemplates_(),
		  projectileTemplates_(),
		  towerEngineData_(new TowerEngineData()),
		  towers_(),
		  enemies_(),
		  projectiles_(),
		  particleSystems_(),
		  particleZ_(0),
		  enemyType_(EnemyData::CalEnemyType),
		  light_(new Light(GL_LIGHT0)),
		  billboardTrue_(true),
		  gameState_(GSWaitingForInput) {
			initDisplay();
			initTemplates();
			initGLSettings();
			initMyStuff();
			initLighting();
			initMatrixes();
			initTextures();
			//initSoftShadows();
			drawSoftShadows_ = false;

			light_->sceneNodeIs(new DrawableSceneNode(light_));
			rootSN_->childNew(light_->sceneNode());
			light_->sceneNode()->translate(5, 0, 3);
		}

		virtual void initMyStuff();
		virtual void onPick(unsigned int pickedId);
		virtual void onKeyDown(SDLKey key, SDLMod modifer);
		virtual void onMouseMotion(unsigned int x, unsigned int y);
		virtual void update(float elapsedMillis);

		Ptr<TowerEngineData> towerEngineData() { return towerEngineData_; }
		Ptr<TowerData> towerData(TowerData::TowerType type) { return towerTemplates_[type]->towerData(); }

		void startRound();
		bool checkRoundOver();

	protected:
		static const int ENEMY_START_ROW;
		static const int ENEMY_START_COL;
		static const unsigned int NUM_ENEMIES;

		map<TowerData::TowerType, Ptr<Tower> > towerTemplates_;
		map<EnemyData::EnemyType, Ptr<Enemy> > enemyTemplates_;
		map<TowerData::TowerType, Ptr<Projectile> > projectileTemplates_;
		Ptr<TowerEngineData> towerEngineData_;
		map<unsigned int, Ptr<Tower> > towers_;
		map<unsigned int, Ptr<Enemy> > enemies_;
		map<unsigned int, Ptr<Projectile> > projectiles_;
		vector<Ptr<ParticleSystem> > particleSystems_;
		float particleZ_;
		EnemyData::EnemyType enemyType_;
		Ptr<Light> light_;
		bool billboardTrue_;
		GameState gameState_;

		void initTemplates();
		void initEnemies(EnemyData::EnemyType enemyType);
		bool isRoundOver();
		bool isGameOver();
		void reset();
		void removeTowers();
		void removeEnemies();
		void removeProjectiles();
		void removeParticles();
		void addTower(Ptr<Tower> tower, unsigned int cellId);
		Ptr<Tower> tower(unsigned int towerId) {
			if (towers_.find(towerId) != towers_.end()) {
				return towers_[towerId];
			}
			return NULL;
		}

		Ptr<Tower> towerNewInstance(TowerData::TowerType towerType);
		Ptr<Enemy> enemyNewInstance(EnemyData::EnemyType enemyType);
		Ptr<Projectile> projectileNewInstance(TowerData::TowerType towerType);
	};
}

#endif